# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bisect
import bitarray as ba
import bitarray.util
from collections import UserDict
from itertools import combinations
import numpy as np
from typing import Generator, Optional, Sequence, Union


from dyada.coordinates import (
    get_coordinates_from_level_index,
    LevelIndex,
    Coordinate,
    coordinate_from_sequence,
    CoordinateInterval,
    location_code_from_float,
    location_code_from_coordinate,
)
from dyada.descriptor import (
    Branch,
    RefinementDescriptor,
    branch_generator,
    int8_ndarray_from_iterable,
)
from dyada.linearization import (
    Linearization,
    MortonOrderLinearization,
    LocationCode,
    bitmask_to_indices,
)


def get_binary_index_from_branch(
    linearization: Linearization, branch: Branch
) -> list[ba.bitarray]:
    num_dimensions = len(branch[0].level_increment)

    # once the branch is found, we can infer the vector index from the branch stack
    current_index_bitarray = [ba.bitarray() for d in range(num_dimensions)]
    history_of_indices, history_of_level_increments = branch.to_history()
    for level_count in range(1, len(branch)):
        bit_index = linearization.get_binary_position_from_index(
            history_of_indices[:level_count],
            history_of_level_increments[:level_count],
        )
        for d in range(num_dimensions):
            d_level_increment = branch[level_count].level_increment[d]
            if d_level_increment > 0:
                current_index_bitarray[d] += bit_index[d : d + 1]

    return current_index_bitarray


def get_level_index_from_branch(
    linearization: Linearization, branch: Branch
) -> LevelIndex:
    num_dimensions = len(branch[0].level_increment)

    current_index_bitarray = get_binary_index_from_branch(linearization, branch)
    current_level = int8_ndarray_from_iterable(
        [len(b) for b in current_index_bitarray],
    )
    current_index = np.fromiter(
        [
            (bitarray.util.ba2int(b) if len(b) > 0 else 0)
            for b in current_index_bitarray
        ],
        dtype=np.int64,
        count=num_dimensions,
    )
    return LevelIndex(
        current_level,
        current_index,
    )


def get_level_index_from_linear_index(
    linearization: Linearization,
    descriptor: RefinementDescriptor,
    linear_index: int,
    is_box_index: bool = True,
) -> LevelIndex:
    current_branch, _ = descriptor.get_branch(linear_index, is_box_index)
    return get_level_index_from_branch(linearization, current_branch)


class Discretization:
    def __init__(self, linearization: Linearization, descriptor: RefinementDescriptor):
        self._linearization = linearization
        self._descriptor = descriptor

    @property
    def descriptor(self):
        return self._descriptor

    def __eq__(self, other):
        return (
            self._linearization == other._linearization
            and self._descriptor == other._descriptor
        )

    def __hash__(self):
        return hash((self._linearization, self._descriptor))

    def __len__(self):
        return self._descriptor.get_num_boxes()

    def __repr__(self):
        return f"Discretization({repr(self._linearization)}, {repr(self._descriptor)})"

    # TODO has pretty dependency
    # def _repr_pretty_(self, p, cycle):
    #    p.text(str(self) if not cycle else '...')

    def __str__(self):
        # this is monkey-patched when drawing.py is imported
        return repr(self)

    def get_level_index_from_branch(self, branch: Branch) -> LevelIndex:
        return get_level_index_from_branch(self._linearization, branch)

    def get_level_index(self, index: int, is_box_index: bool = True) -> LevelIndex:
        return get_level_index_from_linear_index(
            self._linearization, self._descriptor, index, is_box_index
        )

    def get_all_boxes_level_indices(self) -> Generator:
        for i, _ in enumerate(self._descriptor):
            if self._descriptor.is_box(i):
                yield self.get_level_index(i, False)

    def get_containing_box(self, coordinate: Coordinate) -> Union[int, tuple[int, ...]]:
        coordinates_to_find: set[Coordinate] = {tuple(coordinate)}  # type: ignore
        coordinates_found: set[Coordinate] = set()

        found_box_indices: set[int] = set()

        while len(coordinates_to_find - coordinates_found) > 0:
            coordinate = (coordinates_to_find - coordinates_found).pop()
            location_code = location_code_from_coordinate(coordinate)
            found_index, found_part_of_location_code = (
                self.get_index_from_location_code(location_code, get_box=True)  # type: ignore
            )
            found_box_indices.add(found_index)
            coordinates_found.add(coordinate)

            # check if the coordinate could also be in another box
            # this is the case if there are only zeros left
            # in the location_code behind the respective counted levels
            ambiguous_dimensions = [
                location_code[i].count() > 0
                and location_code[i][found_part_of_location_code[i] :].count() == 0
                for i in range(len(coordinate))
            ]
            ambiguous_indices = bitmask_to_indices(ba.bitarray(ambiguous_dimensions))
            for num_ambiguous in range(1, len(ambiguous_indices) + 1):
                for ambiguous_subset in combinations(ambiguous_indices, num_ambiguous):
                    # copy the original coordinate
                    new_coordinate = coordinate_from_sequence(coordinate)  # type: ignore
                    # and subtract a small epsilon from the ambiguous dimensions
                    for i in ambiguous_subset:
                        new_coordinate[i] = np.nextafter(new_coordinate[i], -np.inf)
                    coordinates_to_find.add(tuple(new_coordinate))  # type: ignore

        return (
            tuple(found_box_indices)
            if len(found_box_indices) > 1
            else found_box_indices.pop()
        )

    class NoExactMatchError(Exception):
        def __init__(self, closest_match: int):
            super().__init__()
            self.closest_match = closest_match

    def get_index_from_location_code(
        self, location_code: LocationCode, get_box: bool
    ) -> int | tuple[int, list[int]]:
        branch = Branch(self._descriptor.get_num_dimensions())
        history_of_indices, history_of_refinements = branch.to_history()

        found_bits = [0] * len(location_code)
        descriptor_iter = iter(self._descriptor)

        patch_index = 0
        box_index = -1

        while True:
            refinement = next(descriptor_iter)

            if not get_box and all(
                found_bits[d] >= len(location_code[d])
                for d in range(len(location_code))
            ):
                return patch_index

            branch.grow_branch(refinement)

            if refinement == self._descriptor.d_zeros:
                return (
                    self._handle_leaf(
                        get_box,
                        patch_index,
                        box_index,
                    ),
                    found_bits,
                )

            history_of_refinements.append(refinement)
            binary_position = self._add_refinement_bits(
                refinement,
                location_code,
                found_bits,
                patch_index,
                get_box,
            )

            child_index = self._linearization.get_index_from_binary_position(
                binary_position,
                history_of_indices,
                history_of_refinements,
            )
            history_of_indices.append(child_index)

            patch_index, box_index = self._skip_unmatched_children(
                descriptor_iter,
                branch,
                child_index,
                patch_index,
                box_index,
            )
            patch_index += 1

    def _handle_leaf(self, get_box: bool, patch_index: int, box_index: int):
        if get_box:
            box_index += 1
            return box_index
        raise Discretization.NoExactMatchError(patch_index)

    def _add_refinement_bits(
        self,
        refinement,
        location_code: LocationCode,
        found_bits: list[int],
        patch_index: int,
        get_box: bool,
    ):
        new_binary_position = ba.bitarray(len(location_code))
        for dim, is_refined in enumerate(refinement):
            if not is_refined:
                continue

            bit_pos = found_bits[dim]
            too_short = bit_pos >= len(location_code[dim])
            if too_short and not get_box:
                raise Discretization.NoExactMatchError(patch_index)

            new_binary_position[dim] = 0 if too_short else location_code[dim][bit_pos]
            found_bits[dim] += 1

        return new_binary_position

    def _skip_unmatched_children(
        self,
        descriptor_iter: Generator,
        branch: Branch,
        child_index: int,
        patch_index: int,
        box_index: int,
    ) -> tuple[int, int]:
        for _ in range(child_index):
            skipped = next(descriptor_iter)
            inc_box, inc_patch = self._descriptor.skip_to_next_neighbor(
                descriptor_iter, skipped
            )
            box_index += inc_box
            patch_index += inc_patch
            branch.advance_branch()

        return patch_index, box_index

    def slice(
        self,
        fixed_unit_coordinates: Sequence[Union[float, None]],
        get_level: bool = False,
    ) -> Union[
        tuple["Discretization", dict[int, int]],
        tuple["Discretization", dict[int, int], list[int | None]],
    ]:
        assert isinstance(
            self._linearization, MortonOrderLinearization
        ), "only MortonOrderLinearization is supported (b/c separable),"
        "if you need another one, please imlement the necessary reordering like in refinement"
        assert not all(c is None for c in fixed_unit_coordinates)
        num_dimensions = self._descriptor.get_num_dimensions()
        assert len(fixed_unit_coordinates) == num_dimensions
        fixed_dimensions = [
            i for i, c in enumerate(fixed_unit_coordinates) if c is not None
        ]
        num_fixed_dimensions = len(fixed_dimensions)
        assert num_fixed_dimensions > 0
        assert num_fixed_dimensions < num_dimensions
        maximum_level_in_slice = [
            0 if d in fixed_dimensions else None for d in range(num_dimensions)
        ]
        location_codes = [
            location_code_from_float(c) if c is not None else None
            for c in fixed_unit_coordinates
        ]
        # create new empty descriptor
        new_descriptor = RefinementDescriptor(
            self._descriptor.get_num_dimensions() - num_fixed_dimensions,
        )
        new_descriptor._data = ba.bitarray()
        index_mapping: dict[int, int] = {}

        # iterate the whole descriptor
        max_depth = np.iinfo(np.int16).max
        skip_depth = max_depth
        for current_descriptor_index, (branch, old_refinement) in enumerate(
            branch_generator(self.descriptor)
        ):
            if len(branch) > skip_depth:
                # skip all children
                continue
            else:
                # then continue normal operation
                skip_depth = max_depth
            bitarray_index = get_binary_index_from_branch(self._linearization, branch)
            level = [len(b) for b in bitarray_index]
            keep = True
            for d in fixed_dimensions:
                if level[d] == 0:
                    continue
                # check if the bitarray is equal to the beginnig of the location code
                if not bitarray_index[d] == location_codes[d][: level[d]]:  # type: ignore
                    keep = False
                    break
            if keep:
                # add this refinement to the new descriptor, omitting the fixed dimensions
                new_refinement = ba.bitarray(old_refinement)
                del new_refinement[fixed_dimensions]
                # if the refinement is perpendicular to the fixed dimensions, keep track but don't append
                is_perpendicular = (
                    new_refinement.count() == 0 and old_refinement.count() > 0
                )
                if is_perpendicular:
                    if len(new_descriptor) == 0:
                        index_mapping[current_descriptor_index] = 0
                    else:
                        # look for the parent in the index mapping
                        siblings = self.descriptor.get_siblings(
                            current_descriptor_index
                        )
                        parent_index = siblings[0] - 1
                        index_mapping[current_descriptor_index] = index_mapping[
                            parent_index
                        ]
                else:
                    new_descriptor.get_data().extend(new_refinement)
                    index_mapping[current_descriptor_index] = len(new_descriptor) - 1
                if get_level:
                    maximum_level_in_slice = [
                        (
                            max(maximum_level_in_slice[d], level[d])  # type: ignore
                            if d in fixed_dimensions
                            else None
                        )
                        for d in range(num_dimensions)
                    ]
            else:
                # skip all children as well
                skip_depth = len(branch)
        assert len(new_descriptor) > 0
        if get_level:
            return (
                Discretization(MortonOrderLinearization(), new_descriptor),
                index_mapping,
                maximum_level_in_slice,
            )
        return Discretization(MortonOrderLinearization(), new_descriptor), index_mapping


def coordinates_from_index(
    discretization: Discretization, index: int, is_box_index: bool = False
) -> CoordinateInterval:
    level_index = get_level_index_from_linear_index(
        discretization._linearization,
        discretization._descriptor,
        linear_index=index,
        is_box_index=is_box_index,
    )
    return get_coordinates_from_level_index(level_index)


def coordinates_from_box_index(
    discretization: Discretization,
    index: int,
    full_domain: Optional[CoordinateInterval] = None,
) -> CoordinateInterval:
    coordinates = coordinates_from_index(discretization, index, is_box_index=True)
    if full_domain is not None:
        scaling_factor = full_domain.upper_bound - full_domain.lower_bound
        offset = full_domain.lower_bound
        coordinates = CoordinateInterval(
            lower_bound=coordinates.lower_bound * scaling_factor + offset,
            upper_bound=coordinates.upper_bound * scaling_factor + offset,
        )
    return coordinates


class SliceDictInDimension(UserDict):
    def __init__(
        self, discretization: Discretization, dimension_index: int, box_mapping: bool
    ):
        super().__init__()
        num_dimensions = discretization.descriptor.get_num_dimensions()
        assert dimension_index < num_dimensions
        start = 0.0
        while start < 1.0:
            d_slice, mapping, level_t = discretization.slice(
                [
                    start if d == dimension_index else None
                    for d in range(num_dimensions)
                ],
                get_level=True,  # type: ignore
            )
            if box_mapping:
                mapping = {
                    discretization.descriptor.to_box_index(
                        k
                    ): d_slice.descriptor.to_box_index(v)
                    for k, v in mapping.items()
                    if discretization.descriptor.is_box(k)
                }
            super().__setitem__(start, (d_slice, mapping))
            start += 2.0 ** -int(level_t[dimension_index])  # type: ignore
        self._sorted_keys = sorted(self.data)

    def __setitem__(self, key, value):
        if key < 0.0 or key > 1.0:
            raise KeyError(f"Key {key} out of bounds (0.0 to 1.0)")
        super().__setitem__(key, value)
        bisect.insort(self._sorted_keys, key)

    def __getitem__(self, key):
        if key < 0.0 or key > 1.0:
            raise KeyError(f"Key {key} out of bounds (0.0 to 1.0)")
        idx = bisect.bisect_right(self._sorted_keys, key) - 1
        if idx >= 0:
            return self.data[self._sorted_keys[idx]]
        raise KeyError


def branch_to_location_code(branch: Branch, linearization) -> list[ba.bitarray]:
    """
    Convert a branch to a location code.
    :param branch: The branch to convert.
    :return: A list of bitarrays representing the dimension-separated location code.
    """
    num_dimensions = len(branch[0].level_increment)
    location_code = [ba.bitarray() for _ in range(num_dimensions)]
    assert isinstance(linearization, MortonOrderLinearization)
    for twig in branch:
        refined_dimensions = bitmask_to_indices(twig.level_increment)
        if len(refined_dimensions) == 0:
            continue
        index_from_count = bitarray.util.int2ba(
            (2 ** len(refined_dimensions) - twig.count_to_go_up),
            length=len(refined_dimensions),
        )
        index_from_count.reverse()  # to match our first-dimension-continuous convention
        for i, d_i in enumerate(refined_dimensions):
            location_code[d_i].append(index_from_count[i])
    return location_code


def discretization_to_location_stack_strings(
    discretization: Discretization, plane_symbol="λ"
) -> list[tuple[str, ...]]:
    """
    Create a location stack from a discretization.
    :param discretization: The discretization to convert.
    :return: A list of d-tuples of strings
    """

    location_stack = []
    for current_branch, refinement in branch_generator(discretization.descriptor):
        # update location codes
        location_codes = branch_to_location_code(
            current_branch, discretization._linearization
        )
        # write current location codes to strings and put to stack, append plane_symbol if refinement in this dimension
        location_stack.append(
            tuple(
                location_codes[d].to01() + (plane_symbol if refinement[d] == 1 else "")
                for d in range(len(location_codes))
            )
        )
    return location_stack



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from functools import wraps
import warnings

try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection  # type: ignore
    from matplotlib.colors import to_rgba
except ImportError:
    warnings.warn("matplotlib not found, some plotting functions will not work")

from typing import Sequence, Union, Mapping, Optional

from dyada.coordinates import (
    CoordinateInterval,
    get_coordinates_from_level_index,
)
from dyada.discretization import (
    Discretization,
)
from dyada.drawing_opengl_obj import plot_boxes_3d_pyopengl, export_boxes_3d_to_obj
from dyada.drawing_tikz import plot_boxes_3d_tikz, plot_boxes_2d_tikz
from dyada.drawing_util import (
    labels_from_discretization,
    side_corners_generator,
    boxes_to_2d_ascii,
)
from dyada.structure import depends_on_optional


def plot_boxes_2d(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]] = None,
    projection: Sequence[int] | None = None,
    backend="matplotlib",
    **kwargs,
) -> None:
    if projection is None:
        projection = [0, 1]
    assert len(projection) == 2
    if backend == "matplotlib":
        return plot_boxes_2d_matplotlib(intervals, labels, projection, **kwargs)
    elif backend == "tikz":
        return plot_boxes_2d_tikz(intervals, labels, projection, **kwargs)
    elif backend == "ascii":
        print(boxes_to_2d_ascii(intervals, projection=projection, **kwargs))
    else:
        raise ValueError(f"Unknown backend: {backend}")


def plot_all_boxes_2d(
    discretization: Discretization,
    projection: Sequence[int] | None = None,
    labels: Union[None, str, Sequence[str]] = "patches",
    **kwargs,
) -> None:
    if projection is None:
        projection = [0, 1]
    level_indices = list(discretization.get_all_boxes_level_indices())
    coordinates = [get_coordinates_from_level_index(box_li) for box_li in level_indices]
    labels = labels_from_discretization(discretization, labels)
    plot_boxes_2d(coordinates, projection=projection, labels=labels, **kwargs)


def plot_boxes_3d(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Union[None, str, Sequence[str]] = None,
    projection: Sequence[int] | None = None,
    backend: str = "tikz",
    **kwargs,
) -> None:
    if projection is None:
        projection = [0, 1, 2]
    assert len(projection) == 3
    if backend == "matplotlib":
        return plot_boxes_3d_matplotlib(intervals, labels, projection, **kwargs)
    elif backend == "obj":
        return export_boxes_3d_to_obj(intervals, projection, **kwargs)
    elif backend == "tikz":
        return plot_boxes_3d_tikz(intervals, labels, projection, **kwargs)
    elif backend == "opengl":
        return plot_boxes_3d_pyopengl(intervals, labels, projection, **kwargs)
    else:
        raise ValueError(f"Unknown backend: {backend}")


def plot_all_boxes_3d(
    discretization: Discretization,
    projection: Sequence[int] | None = None,
    labels: Union[None, str, Sequence[str]] = "patches",
    **kwargs,
) -> None:
    if projection is None:
        projection = [0, 1, 2]
    level_indices = list(discretization.get_all_boxes_level_indices())
    coordinates = [get_coordinates_from_level_index(box_li) for box_li in level_indices]
    labels = labels_from_discretization(discretization, labels)
    plot_boxes_3d(coordinates, projection=projection, labels=labels, **kwargs)


def discretization_to_2d_ascii(
    discretization: Discretization,
    resolution=None,
    projection: Sequence[int] | None = None,
    **kwargs,
) -> str:
    if projection is None:
        projection = [0, 1]
    if resolution is None:
        max_level = discretization._descriptor.get_maximum_level()
        resolution = (2 ** (max_level[0] + 1), 2 ** max_level[1])
        return discretization_to_2d_ascii(discretization, resolution)
    level_indices = list(discretization.get_all_boxes_level_indices())
    coordinates = [get_coordinates_from_level_index(box_li) for box_li in level_indices]
    return boxes_to_2d_ascii(
        coordinates, projection=projection, resolution=resolution, **kwargs
    )


def discretization_str(discretization: Discretization):
    dimensionality = discretization._descriptor.get_num_dimensions()
    if dimensionality == 2:
        return discretization_to_2d_ascii(discretization)
    else:
        return repr(discretization)


# Monkey-patch Discretization.__str__ to use this function
Discretization.__str__ = wraps(Discretization.__str__)(discretization_str)  # type: ignore


@depends_on_optional("matplotlib.pyplot")
def get_figure_2d_matplotlib(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    **kwargs,
) -> tuple:
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors = kwargs.pop("colors", prop_cycle.by_key()["color"])
    if isinstance(colors, str):
        colors = [colors] * len(intervals)

    fig, ax1 = plt.subplots(1, 1)
    ax1.set_xlim(0, 1)
    ax1.set_ylim(0, 1)
    for i, interval in enumerate(intervals):
        anchor_point = interval.lower_bound[projection]
        extent = interval.upper_bound[projection] - anchor_point
        rectangle = plt.Rectangle(
            tuple(anchor_point),  # type: ignore
            extent[0],
            extent[1],
            fill=True,
            figure=fig,
            color=colors[i % len(colors)],
            **kwargs,
        )
        ax1.add_artist(rectangle)
        ax1.set_aspect("equal")

        if labels is not None:
            rx, ry = rectangle.get_xy()
            cx = rx + rectangle.get_width() / 2.0
            cy = ry + rectangle.get_height() / 2.0
            ax1.annotate(
                labels[i],
                (cx, cy),
                fontsize=6,
                ha="center",
                va="center",
            )
    # add title with projection
    ax1.set_title(f"Dimensions {projection[0]} and {projection[1]}")
    return fig, ax1


@depends_on_optional("matplotlib.pyplot")
def plot_boxes_2d_matplotlib(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    **kwargs,
) -> None:
    filename = kwargs.pop("filename", None)
    fig, ax1 = get_figure_2d_matplotlib(intervals, labels, projection, **kwargs)
    if filename is not None:
        plt.savefig(filename)
    else:
        plt.show()


@depends_on_optional("matplotlib.pyplot")
def draw_cuboid_on_axis(
    ax: plt.Axes,
    interval: CoordinateInterval,
    projection: Sequence[int],
    color="skyblue",
    wireframe: bool = False,
    **kwargs,
) -> plt.Axes:
    """
    Draw a cuboid on the given axis.
    :param ax: The axis to draw on.
    :param interval: The interval to draw.
    :param projection: The projection to use.
    :param color: The color of the cuboid.
    :param wireframe: Whether to draw the cuboid as a wireframe.
    :param kwargs: Additional arguments to pass to the Poly3DCollection.
    :return: The axis with the cuboid drawn on it.
    """
    lower = interval[0][projection]
    upper = interval[1][projection]
    faces = []
    for side_corners in side_corners_generator(lower, upper):
        face = [*side_corners]
        faces.append(face)
    alpha = kwargs.pop("alpha", 0.5)
    if wireframe:
        color_rgba = to_rgba(color, alpha=alpha)
        cuboid = Poly3DCollection(
            faces,
            facecolors=(0, 0, 0, 0),  # fully transparent faces
            edgecolors=color_rgba,
            **kwargs,
        )
    else:
        edgecolors = kwargs.pop("edgecolors", "gray")
        cuboid = Poly3DCollection(
            faces,
            facecolors=color,
            alpha=alpha,
            **kwargs,
        )
        cuboid.set_edgecolor(edgecolors)
    ax.add_collection(cuboid)
    return ax


@depends_on_optional("matplotlib.pyplot")
def get_figure_3d_matplotlib(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    wireframe: bool = False,
    **kwargs,
) -> tuple:
    if labels is not None:
        warnings.warn("Labels are currently not used in 3D plots w/ matplotlib")

    # plt.ion()
    # plt.show() # using this and the pause below gives a neat animation
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors = kwargs.pop("colors", prop_cycle.by_key()["color"])
    if isinstance(colors, str):
        colors = [colors] * len(intervals)

    fig = plt.figure()
    ax1 = fig.add_subplot(111, projection="3d")
    for i, interval in enumerate(intervals):
        # draw each as cuboid
        draw_cuboid_on_axis(
            ax1,
            interval,
            projection=projection,
            color=colors[i % len(colors)],
            wireframe=wireframe,
            **kwargs,
        )
        # plt.pause(0.01)

    # add title with projection
    ax1.set_title(f"Dimensions {projection[0]}, {projection[1]}, {projection[2]}")
    return fig, ax1


@depends_on_optional("matplotlib.pyplot")
def plot_boxes_3d_matplotlib(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    **kwargs,
) -> None:
    filename = kwargs.pop("filename", None)
    fig, ax1 = get_figure_3d_matplotlib(intervals, labels, projection, **kwargs)
    if filename is not None:
        plt.savefig(filename)
    else:
        plt.show()



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
from collections import defaultdict
from dataclasses import dataclass
import numpy as np
from typing import TypeAlias, Union

from dyada.coordinates import bitarray_startswith
from dyada.descriptor import (
    RefinementDescriptor,
    Branch,
)
from dyada.discretization import Discretization, branch_to_location_code
from dyada.linearization import (
    CoarseningTracker,
    TrackToken,
    LocationCode,
    binary_or_none_generator,
    bitmask_to_indices,
    get_initial_coarsen_refine_stack,
    location_code_from_history,
    location_code_from_branch,
)
from dyada.markers import MarkerType, MarkersMapProxyType


class AncestryBranch:
    """An AncestryBranch couples a Branch together with a history of binary positions
     and node indices (as in the descriptors get_ancestry function).

    It can be used to iterate a new descriptor before it's explicitly constructed, by
    using the markers to determine to which location / position in the old discretization to grow next.

    Invariants:
    - branch, history_of_indices, history_of_level_increments, history_of_binary_positions and ancestry
      are always consistent with each other
    - ancestry corresponds to the old indices on the current_modified_branch, but with potentially
      new relationships (not determined by old descriptor)
    """

    TrackInfo: TypeAlias = CoarseningTracker

    @dataclass
    class _AncestryBranchState:
        discretization: Discretization
        current_modified_branch: Branch
        initial_branch_depth: int
        history_of_indices: list[int]
        history_of_level_increments: list[ba.frozenbitarray]
        history_of_binary_positions: list[ba.bitarray]

        def __init__(self, discretization: Discretization, starting_index: int):
            self.discretization = discretization
            descriptor = discretization.descriptor
            self.current_modified_branch, _ = descriptor.get_branch(
                starting_index, is_box_index=False
            )
            self.initial_branch_depth = len(self.current_modified_branch)

            self.history_of_indices, self.history_of_level_increments = (
                self.current_modified_branch.to_history()
            )
            self.history_of_binary_positions = [
                discretization._linearization.get_binary_position_from_index(
                    self.history_of_indices[: i + 1],
                    self.history_of_level_increments[: i + 1],
                )
                for i in range(len(self.current_modified_branch) - 1)
            ]

        def __len__(self) -> int:
            return len(self.history_of_binary_positions)

        def advance(self) -> int:
            self.current_modified_branch.advance_branch(self.initial_branch_depth)

            # prune all other data to current length,
            # so we don't have to recompute from new branch
            current_modified_branch_depth = len(self.current_modified_branch)
            self.history_of_binary_positions = self.history_of_binary_positions[
                : current_modified_branch_depth - 2
            ]
            self.history_of_indices = self.history_of_indices[
                : current_modified_branch_depth - 1
            ]
            self.history_of_indices[-1] += 1
            self.history_of_level_increments = self.history_of_level_increments[
                : current_modified_branch_depth - 1
            ]

            # update history of binary positions
            latest_binary_position = (
                self.discretization._linearization.get_binary_position_from_index(
                    self.history_of_indices, self.history_of_level_increments
                )
            )
            self.history_of_binary_positions.append(latest_binary_position)
            return current_modified_branch_depth

        def grow(self, next_refinement: ba.frozenbitarray) -> None:
            self.current_modified_branch.grow_branch(next_refinement)
            self.history_of_level_increments.append(next_refinement)
            self.history_of_indices.append(0)
            # update history of binary positions
            latest_binary_position = (
                self.discretization._linearization.get_binary_position_from_index(
                    self.history_of_indices, self.history_of_level_increments
                )
            )
            self.history_of_binary_positions.append(latest_binary_position)

    @dataclass
    class _AncestryMappingState:
        # the ancestry, in old indices but new relationships
        ancestry: list[int]
        old_indices_map_track_tokens: defaultdict[int, list[TrackToken]]
        current_track_token: TrackToken
        missed_mappings: defaultdict[int, set[TrackToken]]
        track_info_mapping: dict[int, "AncestryBranch.TrackInfo"]

        def __init__(
            self,
            discretization: Discretization,
            ancestry_branch_state: "AncestryBranch._AncestryBranchState",
        ):
            self.ancestry = discretization.descriptor.get_ancestry(
                ancestry_branch_state.current_modified_branch
            )
            self.old_indices_map_track_tokens = defaultdict(list)
            self.current_track_token = TrackToken(-1)
            self.missed_mappings = defaultdict(set)
            self.track_info_mapping = {}

        def update(
            self,
            discretization,
            current_old_index: int,
            markers: MarkersMapProxyType,
            next_marker: MarkerType,
            intermediate_generation: set[int],
            exact: bool,
        ) -> TrackToken:
            self.current_track_token = TrackToken(self.current_track_token.t + 1)

            self.store_missed_mappings(
                current_old_index,
                intermediate_generation,
                exact,
            )
            # process new track info
            if current_old_index not in self.track_info_mapping:
                relevant_markers = [next_marker]
                for ancestor_index in self.ancestry:
                    relevant_markers.append(markers[ancestor_index])
                most_recent_track_info = self._get_initial_track_info(
                    discretization.descriptor[current_old_index],
                    relevant_markers,
                )
                if most_recent_track_info is not None:
                    self.track_info_mapping[current_old_index] = most_recent_track_info
            self.ancestry.append(current_old_index)
            self.old_indices_map_track_tokens[current_old_index].append(
                self.current_track_token
            )

            return self.current_track_token

        def _get_initial_track_info(
            self,
            current_refinement: ba.frozenbitarray,
            markers: list[MarkerType],
        ) -> Union["AncestryBranch.TrackInfo", None]:
            if current_refinement.count() == 0:
                return None
            marker = np.sum(markers, axis=0)
            if marker.min() < 0:
                negative = ba.bitarray(
                    1 if marker[i] < 0 else 0 for i in range(len(marker))
                )

                dimensions_to_coarsen = ba.frozenbitarray(negative & current_refinement)
                dimensions_cannot_coarsen = ba.frozenbitarray(
                    negative & ~current_refinement
                )
                dimensions_to_refine = ba.frozenbitarray(
                    ba.frozenbitarray(
                        1 if marker[i] > 0 else 0 for i in range(len(marker))
                    )
                )
                coarsen_refine_stack = get_initial_coarsen_refine_stack(
                    ba.frozenbitarray(current_refinement),
                    dimensions_to_coarsen,
                    dimensions_to_refine,
                    (
                        dimensions_cannot_coarsen
                        if dimensions_cannot_coarsen.count() > 0
                        else None
                    ),
                )
                return coarsen_refine_stack
            return None

        def store_missed_mappings(
            self, current_old_index: int, intermediate_generation: set[int], exact: bool
        ) -> None:
            if not self.ancestry:
                return  # at root, nothing to map
            parent_index = self.ancestry[-1]
            parent_tokens = self.old_indices_map_track_tokens[parent_index]

            for skipped_index in intermediate_generation:
                if skipped_index < current_old_index:  # is ancestor
                    # map upward to parent
                    self.missed_mappings[skipped_index].add(parent_tokens[-1])
                # either descendant -> map upward to current, or ancestor -> map downward to current
                self.missed_mappings[skipped_index].add(self.current_track_token)

            if not exact:
                # no exact match, map current_old_index upward
                self.missed_mappings[current_old_index].add(parent_tokens[-1])
                for ancestor_index, ancestor in reversed(
                    list(enumerate(self.ancestry))
                ):
                    ancestor_exact = len(self.missed_mappings.get(ancestor, [])) == 0
                    if ancestor_exact:
                        break
                    older_ancestor = self.ancestry[ancestor_index - 1]
                    older_ancestor_token = min(
                        self.old_indices_map_track_tokens[older_ancestor]
                    )
                    self.missed_mappings[current_old_index].add(older_ancestor_token)
            # update existing track info
            ancestor_track_info = self.track_info_mapping.get(parent_index)
            if ancestor_track_info is not None:
                local_position, same_index_as = ancestor_track_info.pop()
                if same_index_as is None:
                    ancestor_track_info.register_group(
                        local_position, self.current_track_token
                    )

    def __init__(
        self,
        discretization: Discretization,
        starting_index: int,
        markers: MarkersMapProxyType,
    ):
        self.markers = markers
        self._discretization = discretization
        # iterates a modified version of the descriptor, incorporating the markers knowledge
        # and keeping track of the ancestry
        self._branch_state = AncestryBranch._AncestryBranchState(
            discretization, starting_index
        )
        self._mapping_state = AncestryBranch._AncestryMappingState(
            discretization, self._branch_state
        )
        assert (
            len(self._mapping_state.ancestry)
            == self._branch_state.initial_branch_depth - 1
        )

    def get_current_location_info(
        self,
    ) -> tuple[int, TrackToken, ba.frozenbitarray, MarkerType]:
        # get the currently desired location info
        current_old_index = 0
        intermediate_generation: set[int] = set()
        exact = True
        if len(self._branch_state) > 0:  # if not at root
            modified_dimensionwise_positions = location_code_from_history(
                self._branch_state.history_of_binary_positions,
                self._branch_state.history_of_level_increments,
            )
            current_old_index, intermediate_generation, exact = _find_next_twig(
                self._discretization,
                self.markers,
                modified_dimensionwise_positions,
                self._mapping_state.ancestry[-1],
            )
        next_refinement = _refinement_with_marker_applied(
            self._discretization.descriptor[current_old_index],
            next_marker := self.markers[current_old_index],
        )

        current_track_token = self._mapping_state.update(
            self._discretization,
            current_old_index,
            self.markers,
            next_marker,
            intermediate_generation,
            exact,
        )
        return (
            current_old_index,
            current_track_token,
            next_refinement,
            next_marker,
        )

    class WeAreDoneAndHereAreTheMissingRelationships(Exception):
        def __init__(
            self,
            mapping: dict[int, set[TrackToken]],
        ):
            self.missing_mapping = mapping
            super().__init__()

    def advance(self) -> None:
        try:
            current_modified_branch_depth = self._branch_state.advance()
        except IndexError as e:
            mapping = self._gather_remaining_mappings()
            raise AncestryBranch.WeAreDoneAndHereAreTheMissingRelationships(
                mapping
            ) from e

        self._mapping_state.ancestry = self._mapping_state.ancestry[
            : current_modified_branch_depth - 1
        ]

    def grow(self, next_refinement: ba.frozenbitarray) -> None:
        self._branch_state.grow(next_refinement)

    def _gather_remaining_mappings(self) -> dict[int, set[TrackToken]]:
        # check if all relationships from coarsening tracking are exhausted
        mapping: defaultdict[int, set[TrackToken]] = self._mapping_state.missed_mappings
        hint_previous_branch = None
        dimensionality = self._discretization.descriptor.get_num_dimensions()
        for key, track_info in sorted(self._mapping_state.track_info_mapping.items()):
            ancestor_branch = self._discretization.descriptor.get_branch(
                key, is_box_index=False, hint_previous_branch=hint_previous_branch
            )[0]
            hint_previous_branch = (key, ancestor_branch)
            ancestor_location_code = branch_to_location_code(
                ancestor_branch, self._discretization._linearization
            )
            unresolved_coarsen_mask = track_info.unresolved_mask
            unresolved_iterable = None
            if unresolved_coarsen_mask is not None:
                # shorten the ancestor location code for each unresolved coarsen dimension
                for d, b in enumerate(unresolved_coarsen_mask):
                    if b:
                        ancestor_location_code[d].pop()
                unresolved_iterable = binary_or_none_generator(
                    bitmask_to_indices(unresolved_coarsen_mask),
                    dimensionality,
                )
                # todo get analytically
                num_to_drop = 2**dimensionality - len(track_info)
                for _ in range(num_to_drop):
                    next(unresolved_iterable)
            current_refinement_dimensions = bitmask_to_indices(
                self._discretization.descriptor[key]
            )
            while len(track_info) > 0:
                local_position, same_index_as = track_info.pop()
                # get their indices in the old discretization by their location code
                missed_descendant_location_code = [
                    a.copy() for a in ancestor_location_code
                ]
                for d in current_refinement_dimensions:
                    if unresolved_iterable is not None:
                        unresolved_to_append = next(unresolved_iterable)
                        for ud in range(len(unresolved_to_append)):
                            if unresolved_to_append[ud] is None:
                                continue
                            missed_descendant_location_code[ud].append(
                                unresolved_to_append[ud]
                            )
                    missed_descendant_location_code[d].append(local_position[d])

                missed_descendant_index = self._discretization.get_index_from_location_code(
                    missed_descendant_location_code, get_box=False  # type: ignore
                )
                if same_index_as is None:
                    map_to = self._mapping_state.old_indices_map_track_tokens[key]
                else:
                    map_to = [same_index_as]

                assert isinstance(missed_descendant_index, int)
                mapping[missed_descendant_index].update(map_to)
            assert (  # assert exhausted
                unresolved_iterable is None or next(unresolved_iterable, None) is None
            )

        return mapping


def _refinement_with_marker_applied(
    refinement: ba.frozenbitarray,
    marker: MarkerType,
) -> ba.frozenbitarray:
    if refinement.count() == 0:
        assert np.all(marker > -1)
        return refinement

    positive = ba.bitarray([1 if m > 0 else 0 for m in marker])
    assert (refinement & positive).count() == 0
    refinement_modified = refinement | positive

    negative = ba.bitarray([1 if m < 0 else 0 for m in marker])
    assert (~refinement & negative).count() == 0
    refinement_modified ^= negative
    return ba.frozenbitarray(refinement_modified)


def _is_old_index_now_at_or_containing_location_code(
    discretization: Discretization,
    markers: MarkersMapProxyType,
    desired_dimensionwise_positions: LocationCode,
    parent_of_next_refinement: int,
    parent_branch: Branch,
    old_index: int,
) -> tuple[bool, LocationCode]:
    """Check whether old_index is now at or containing the next hyperrectangular location code.
    Args:
        discretization (Discretization): the old discretization we're referring to
        markers (MarkersMapProxyType): markers that should be applied to the old discretization
        desired_dimensionwise_positions (LocationCode): the location code we're looking for next
        parent_of_next_refinement (int): the current parent or other ancestor of the index we're looking for
        parent_branch (Branch): the branch corresponding to the parent_of_next_refinement
        old_index (int): the old index we're checking

    Returns:
        bool         : whether old_index is now at or containing the location code
        LocationCode : the location code of the old_index
    """
    descriptor = discretization.descriptor
    old_index_branch, _ = descriptor.get_branch(
        old_index,
        is_box_index=False,
        hint_previous_branch=(parent_of_next_refinement, parent_branch),
    )
    old_index_dimensionwise_positions = location_code_from_branch(
        old_index_branch, discretization._linearization
    )

    old_index_ancestors = descriptor.get_ancestry(old_index_branch)
    old_index_ancestry_accumulated_markers = np.sum(
        [markers[ancestor] for ancestor in old_index_ancestors],
        axis=0,
    )
    shortened_parent_positions = [
        desired_dimensionwise_positions[d][
            : len(desired_dimensionwise_positions[d])
            - old_index_ancestry_accumulated_markers[d]
        ]
        for d in range(descriptor.get_num_dimensions())
    ]
    part_of_history = all(
        bitarray_startswith(
            old_index_dimensionwise_positions[d],
            shortened_parent_positions[d],
        )
        for d in range(descriptor.get_num_dimensions())
    )
    return part_of_history, old_index_dimensionwise_positions


def _old_node_will_be_in_new_descriptor(
    descriptor: RefinementDescriptor,
    old_index: int,
    markers: MarkersMapProxyType,
) -> bool:
    """Check whether the node at old_index will be contained in the new descriptor after applying markers.

    Returns:
        bool: True if the node will be contained, False if it will be coarsened away.
    """
    future_refinement = _refinement_with_marker_applied(
        descriptor[old_index], marker := markers[old_index]
    )
    return np.min(marker) >= 0 or future_refinement != descriptor.d_zeros  # type: ignore


def _find_next_twig(
    discretization: Discretization,
    markers: MarkersMapProxyType,
    desired_dimensionwise_positions: LocationCode,
    parent_of_next_refinement: int,
) -> tuple[int, set[int], bool]:
    """Get the (old) tree node corresponding to the location code, and any nodes encountered on the way.
    Args:
        discretization (Discretization): the old discretization we're referring to
        markers (MarkersMapProxyType): refinement markers that should be applied to the discretization
        desired_dimensionwise_positions (list[ba.bitarray]): the location code we're looking for next
        parent_of_next_refinement (int): parent or other ancestor of the index we're looking for
    Returns:
        tuple[int, set[int], bool]:
            tuple consisting of the (old) node index
            and a set of (otherwise forgotten) intermediate nodes
            and a bool indicating whether the node is exactly at the location code
    """
    descriptor = discretization.descriptor
    parent_branch, _ = descriptor.get_branch(parent_of_next_refinement, False)
    children = descriptor.get_children(parent_of_next_refinement, parent_branch)
    intermediate_generation: set[int] = set()
    while True:
        for child in children:
            part_of_history, child_dimensionwise_positions = (
                _is_old_index_now_at_or_containing_location_code(
                    discretization,
                    markers,
                    desired_dimensionwise_positions,
                    parent_of_next_refinement,
                    parent_branch,
                    child,
                )
            )
            if not part_of_history:
                continue

            location_code_matches = (
                child_dimensionwise_positions == desired_dimensionwise_positions
            )
            if _old_node_will_be_in_new_descriptor(descriptor, child, markers):
                # we found the next twig
                return child, intermediate_generation, location_code_matches

            # else it's a coarsened node and we can see if there is a matching child
            children_of_coarsened = descriptor.get_children(child)
            if location_code_matches:
                # this means that its former children are now gone and need to be mapped to this child's index
                for child_of_coarsened in children_of_coarsened:
                    # need to append the binarized index of the child, broadcast to split dimensions
                    intermediate_generation.add(child_of_coarsened)
                return child, intermediate_generation, True
            else:
                # else, it's an ancestor node that's going to disappear
                # -> restart loop with new children and remember this one
                intermediate_generation.add(child)
                children = children_of_coarsened



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import bitarray as ba
from collections import deque
import numpy as np
from os.path import abspath

from dyada.descriptor import (
    LevelCounter,
    DyadaInvalidDescriptorError,
    RefinementDescriptor,
    validate_descriptor,
)

from dyada.refinement import Discretization, apply_single_refinement
from dyada.linearization import MortonOrderLinearization


def test_construct():
    for i in range(1, 128):
        r = RefinementDescriptor(i)
        assert r
        validate_descriptor(r)
        assert repr(r).startswith("RefinementDescriptor")


def test_invalid_refinement():
    with pytest.raises(ValueError):
        RefinementDescriptor.from_binary(3, ba.bitarray("01010"))
    with pytest.raises(ValueError):
        RefinementDescriptor.from_binary(3, ba.bitarray("010101"))
    r = RefinementDescriptor(3)
    r._data = ba.bitarray("010101")  # manually corrupt
    with pytest.raises(DyadaInvalidDescriptorError):
        validate_descriptor(r)
    r._data = ba.bitarray("01010")  # manually corrupt
    with pytest.raises(DyadaInvalidDescriptorError):
        validate_descriptor(r)
    r = RefinementDescriptor(2)
    r._data = ba.bitarray("01 00 01 00 00 01 00 00")  # manually corrupt
    with pytest.raises(DyadaInvalidDescriptorError):
        validate_descriptor(r)


def test_zero_level():
    for i in range(1, 10):
        r = RefinementDescriptor(i, 0)
        assert len(r) == 1
        assert not r.get_data().any()
        assert r.get_num_boxes() == 1
        validate_descriptor(r)


def test_one_level():
    for i in range(1, 2):
        r = RefinementDescriptor(i, 1)
        d = r.get_num_dimensions()
        assert d == i
        assert len(r) == 2**i + 1
        assert r.get_num_boxes() == 2**i
        assert r.get_data().count() == i
        assert r.get_data()[0] == 1
        assert r.is_pow2tree()
        validate_descriptor(r)


def test_six_d():
    for level in range(1, 4):
        r = RefinementDescriptor(6, level)
        assert r.get_num_dimensions() == 6
        acc = 1
        for i in range(level):
            acc = acc * 2**6 + 1
        assert len(r) == acc
        assert r.get_num_boxes() == 2 ** (level * 6)
        # count length of one-blocks
        lengths = []
        current_length = 0
        for i in range(0, len(r.get_data()), 6):
            c = r.get_data()[i : i + 6].count()
            # either all-zero or all-one
            assert c in [0, 6]
            if c == 6:
                current_length += 1
            elif current_length != 0:
                lengths.append(current_length)
                current_length = 0
        # assert lengths == generalized_ruler(6, level - 1).tolist()
        assert r.is_pow2tree()
        validate_descriptor(r)


def test_construct_anisotropic():
    r = RefinementDescriptor(4, [0, 1, 2, 3])
    assert r.get_num_dimensions() == 4
    assert len(r) == ((2 + 1) * 4 + 1) * 8 + 1
    assert r.get_num_boxes() == 2**6
    assert r.get_data()[: 4 * 5] == ba.bitarray("01110011000100000000")
    assert r.get_data()[4 * -6 :] == ba.bitarray("000100000000000100000000")
    assert r[:5] == r.get_data()[: 4 * 5]
    assert r[5] == r.get_data()[4 * 5 : 4 * 6]
    assert r[-6:] == r.get_data()[4 * -6 :]
    assert not r.is_pow2tree()
    validate_descriptor(r)


def test_get_level_isotropic():
    r = RefinementDescriptor(4, 3)
    with pytest.raises(IndexError):
        r.get_level(len(r), False)
    with pytest.raises(IndexError):
        r.get_level(-1, False)
    assert np.array_equal(r.get_level(0, False), [0, 0, 0, 0])
    assert np.array_equal(r.get_level(1, False), [1, 1, 1, 1])
    assert np.array_equal(r.get_level(2, False), [2, 2, 2, 2])
    for i in range(0, 16):
        assert np.array_equal(r.get_level(3 + i, False), [3, 3, 3, 3])
        assert np.array_equal(r.get_level(len(r) - i - 1, False), [3, 3, 3, 3])
    assert np.array_equal(r.get_level(len(r) - 16 - 1, False), [2, 2, 2, 2])


def test_get_branch():
    r = RefinementDescriptor(2, [1, 2])
    assert r.get_branch(2, False)[0] == r.get_branch(0, True)[0]
    assert r.get_branch(3, False)[0] == r.get_branch(1, True)[0]
    assert r.get_branch(5, False)[0] == r.get_branch(2, True)[0]
    assert r.get_branch(6, False)[0] == r.get_branch(3, True)[0]
    assert r.get_branch(8, False)[0] == r.get_branch(4, True)[0]
    assert r.get_branch(9, False)[0] == r.get_branch(5, True)[0]
    assert r.get_branch(9, False)[0] == deque(
        [
            LevelCounter(
                level_increment=ba.frozenbitarray("00"),
                count_to_go_up=1,
            ),
            LevelCounter(
                level_increment=ba.frozenbitarray("11"),
                count_to_go_up=2,
            ),
            LevelCounter(
                level_increment=ba.frozenbitarray("01"),
                count_to_go_up=1,
            ),
        ]
    )
    # test the branch's repr
    assert (repr(r.get_branch(9, False)[0])).startswith("Branch")


def test_to_box_index():
    # according to the same mapping as in test_get_level_index
    r = RefinementDescriptor(2, [1, 2])
    for non_box_index in (0, 1, 4, 7, 10, 13):
        with pytest.raises(AssertionError):
            r.to_box_index(non_box_index)
    assert r.to_box_index(2) == 0
    assert r.to_box_index(3) == 1
    assert r.to_box_index(5) == 2
    assert r.to_box_index(6) == 3
    assert r.to_box_index(8) == 4
    assert r.to_box_index(9) == 5
    assert r.to_box_index(11) == 6
    assert r.to_box_index(12) == 7


def test_to_hierarchical_index():
    # according to the same mapping as in test_get_level_index
    r = RefinementDescriptor(2, [1, 2])
    assert r.to_hierarchical_index(0) == 2
    assert r.to_hierarchical_index(1) == 3
    assert r.to_hierarchical_index(2) == 5
    assert r.to_hierarchical_index(3) == 6
    assert r.to_hierarchical_index(4) == 8
    assert r.to_hierarchical_index(5) == 9
    assert r.to_hierarchical_index(6) == 11
    assert r.to_hierarchical_index(7) == 12
    with pytest.raises(AssertionError):
        r.to_hierarchical_index(8)


def test_family_relations():
    r = RefinementDescriptor(2, [1, 2])

    families = {
        -1: [0],
        0: [1, 4, 7, 10],
        1: [2, 3],
        4: [5, 6],
        7: [8, 9],
        10: [11, 12],
    }
    for parent, siblings in families.items():
        for sibling in siblings:
            sibling_branch = r.get_branch(sibling, False)[0]
            assert r.get_parent(sibling_branch)[0] == parent
            assert r.get_siblings(sibling) == siblings
        assert r.get_children(parent) == siblings
    for leaf in (2, 3, 5, 6, 8, 9, 11, 12):
        assert r.get_children(leaf) == []

    with pytest.raises(IndexError):
        r.get_siblings(13)


def test_write_read():
    r = RefinementDescriptor(3, [5, 1, 2])
    r.to_file("test_write")

    r_read = RefinementDescriptor.from_file("test_write_3d.bin")
    assert r == r_read


def test_get_maximum_level_randomized():
    for d in range(1, 6):
        r = RefinementDescriptor(d, 2)
        # have a random refinement
        refinement = np.random.randint(
            0,
            4,
            size=d,
        )
        # apply to a random index in the discretization
        discretization = Discretization(MortonOrderLinearization(), r)
        index = np.random.randint(0, len(discretization))
        new_discretization, _ = apply_single_refinement(
            discretization,
            index,
            refinement,
        )

        max_level = new_discretization.descriptor.get_maximum_level()
        assert all(max_level == refinement + 2 * np.ones(d, dtype=int))


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from itertools import product
import bitarray as ba
from dataclasses import dataclass
from typing import Sequence, TypeAlias

from dyada.structure import copying_lru_cache


def single_bit_set_gen(num_dimensions: int):
    for i in range(num_dimensions):
        bit_array = ba.bitarray(num_dimensions)
        bit_array[i] = 1
        yield bit_array


class Linearization(ABC):
    @staticmethod
    @abstractmethod
    def get_binary_position_from_index(
        history_of_indices: Sequence[int],
        history_of_level_increments: Sequence[ba.bitarray],
    ) -> ba.bitarray: ...

    @staticmethod
    @abstractmethod
    def get_index_from_binary_position(
        binary_position: ba.bitarray,
        history_of_indices: Sequence[int],
        history_of_level_increments: Sequence[ba.bitarray],
    ) -> int: ...

    def __eq__(self, other):
        if not isinstance(other, Linearization):
            return False
        # for now, equality is just type equality
        return type(self) is type(other)

    def __hash__(self):
        return hash(type(self))


class MortonOrderLinearization(Linearization):
    @staticmethod
    def get_binary_position_from_index(
        history_of_indices: Sequence[int],
        history_of_level_increments: Sequence[ba.bitarray],
    ) -> ba.bitarray:
        assert len(history_of_indices) == len(history_of_level_increments)

        this_level_increment = history_of_level_increments[-1]
        assert this_level_increment.count() > 0
        index_in_box = history_of_indices[-1]
        if not index_in_box < 2 ** this_level_increment.count() or index_in_box < 0:
            raise IndexError("Index " + str(index_in_box) + " out of bounds")

        number_of_dimensions = len(this_level_increment)
        binary_position = ba.bitarray(number_of_dimensions)
        # first dimension is the most contiguous
        for dim_index in range(number_of_dimensions):
            if this_level_increment[dim_index]:
                binary_position[dim_index] = index_in_box & 1
                index_in_box >>= 1
        return binary_position

    @staticmethod
    def get_index_from_binary_position(
        binary_position: ba.bitarray,
        history_of_indices: Sequence[int],
        history_of_level_increments: Sequence[ba.bitarray],
    ) -> int:
        assert len(history_of_indices) == len(history_of_level_increments) - 1

        this_level_increment = history_of_level_increments[-1]
        assert this_level_increment.count() > 0
        assert len(binary_position) == len(this_level_increment)
        for i in range(len(binary_position)):
            if binary_position[i]:
                assert this_level_increment[i]
        number_of_dimensions = len(this_level_increment)
        index_in_box = 0

        # first dimension is the most contiguous
        for dim_index in reversed(range(number_of_dimensions)):
            if this_level_increment[dim_index]:
                index_in_box <<= 1
                if binary_position[dim_index]:
                    index_in_box += 1

        return index_in_box


LocationCode: TypeAlias = tuple[ba.frozenbitarray, ...]


def location_code_from_history(
    history_of_binary_positions: Sequence[ba.bitarray],
    history_of_level_increments: Sequence[ba.bitarray],
) -> LocationCode:
    num_dimensions = len(history_of_binary_positions[0])
    depth = len(history_of_binary_positions)
    assert depth > 0
    assert len(history_of_level_increments) == depth
    transposed_positions = [
        ba.bitarray([position[d] for position in history_of_binary_positions])
        for d in range(num_dimensions)
    ]
    transposed_level_increments = [
        ba.bitarray([increment[d] for increment in history_of_level_increments])
        for d in range(num_dimensions)
    ]
    return tuple(
        ba.frozenbitarray(transposed_positions[d][transposed_level_increments[d]])
        for d in range(num_dimensions)
    )


def location_code_from_branch(branch, linearization: Linearization) -> LocationCode:
    history_of_indices, history_of_level_increments = branch.to_history()
    depth = len(history_of_indices)
    assert len(history_of_level_increments) == depth
    history_of_binary_positions = []
    for i in range(depth):
        history_of_binary_positions.append(
            linearization.get_binary_position_from_index(
                history_of_indices[: i + 1],
                history_of_level_increments[: i + 1],
            )
        )
    return location_code_from_history(
        history_of_binary_positions, history_of_level_increments
    )


def binary_or_none_generator(indices: Sequence[int], N: int):
    """Generate all tuples of length N with 0/1 at given indices, None elsewhere.
    None also repeats.

    Args:
        indices (Sequence[int]): indices for which to yield 0/1
        N (int): total number of positions

    Yields:
        tuple[Union[int, None]]: tuples of length N with 0/1 at given indices, None elsewhere
    """
    # value indices reversed so lower index flips faster
    value_indices = list(reversed(sorted(indices)))
    none_indices = [i for i in range(N) if i not in indices]

    pools = [[0, 1]] * len(value_indices) + [[None, None]] * len(none_indices)

    for vals in product(*pools):
        out: list[int | None] = [None] * N

        for i, v in zip(value_indices, vals[: len(value_indices)]):
            out[i] = v

        yield tuple(out)


@dataclass(frozen=True, order=True)
class TrackToken:
    t: int


class ChildGroupingTracker:
    """Groups children of a parent node by a bitmask partition.

    Children are sorted by `sort_mask` bits, so entries that share the same
    `_group_mask` bits (the complement of `sort_mask` and `unresolved_mask`)
    are adjacent in pop order. `pop()` and `register_group()` use
    `_group_mask` to determine group membership.

    Used by coarsening (sort_mask = dims being removed, group by remaining dims)
    and downsplit (sort_mask = dims kept at parent, group by pushed-down dims).
    """

    def __init__(
        self,
        sort_mask: ba.frozenbitarray,
        unresolved_mask: ba.frozenbitarray | None,
        positions: list[ba.frozenbitarray],
    ):
        self.sort_mask = sort_mask
        self.unresolved_mask = unresolved_mask
        self._group_mask = ba.frozenbitarray(
            ~sort_mask & ~unresolved_mask if unresolved_mask else ~sort_mask
        )
        self._group_pointers: dict[ba.frozenbitarray, TrackToken] = {}
        self._positions = positions  # reversed, so pop() gives correct order

    # Keep old attribute names accessible for existing code
    @property
    def separated_mask(self) -> ba.frozenbitarray:
        return self.sort_mask

    def __len__(self) -> int:
        return len(self._positions)

    def pop(self) -> tuple[ba.frozenbitarray, TrackToken | None]:
        """Pop next position. Returns (local_position, same_group_as)."""
        pos = self._positions.pop()
        group_key = ba.frozenbitarray(pos[self._group_mask])
        return pos, self._group_pointers.get(group_key)

    def register_group(
        self, local_position: ba.frozenbitarray, token: TrackToken
    ) -> None:
        """Register that a group maps to the given token."""
        group_key = ba.frozenbitarray(local_position[self._group_mask])
        self._group_pointers[group_key] = token


def indices_to_bitmask(
    indices: Sequence[int], num_dimensions: int
) -> ba.frozenbitarray:
    return ba.frozenbitarray(1 if i in indices else 0 for i in range(num_dimensions))


def bitmask_to_indices(
    bitmask: ba.bitarray | ba.frozenbitarray,
) -> tuple[int, ...]:
    return tuple(i for i in range(len(bitmask)) if bitmask[i])


@copying_lru_cache()
def get_initial_child_grouping(
    current_parent_refinement: ba.frozenbitarray,
    sort_dimensions: ba.frozenbitarray,
    unresolved_dimensions: ba.frozenbitarray | None = None,
    linearization: Linearization = MortonOrderLinearization(),
) -> ChildGroupingTracker:
    """Returns a ChildGroupingTracker for all current children of a parent patch.

    Children are sorted by `sort_dimensions` bits, so entries sharing the same
    sort-dimension values are adjacent. Groups are distinguished by the
    complementary bits.

    For coarsening: sort_dimensions = dims being removed (children to merge
    are adjacent). For downsplit: sort_dimensions = dims kept at parent
    (children in the same target-child group are adjacent).

    Args:
        current_parent_refinement: current parent's refinement, frozen to be hashable
        sort_dimensions: bitmask of dimensions that determine the sort order
        unresolved_dimensions: bitmask of dimensions that cannot be resolved
        linearization: Linearization. Defaults to MortonOrderLinearization().
    Returns:
        ChildGroupingTracker with positions reversed so pop() gives correct order.
    """
    if not isinstance(linearization, MortonOrderLinearization):
        raise NotImplementedError(
            "Initial child grouping only implemented for "
            "Morton order linearization, other linearizations may need "
            "different signature."
        )
    assert len(sort_dimensions) == len(current_parent_refinement)
    assert (sort_dimensions & ~current_parent_refinement).count() == 0
    if unresolved_dimensions is not None:
        current_parent_refinement = ba.frozenbitarray(
            current_parent_refinement | unresolved_dimensions
        )
    positions: list[ba.frozenbitarray] = []
    num_current_children = 2 ** current_parent_refinement.count()
    for child_index in range(num_current_children):
        binary_position = linearization.get_binary_position_from_index(
            (child_index,),
            (current_parent_refinement,),
        )
        positions.append(ba.frozenbitarray(binary_position))

    def _sort_key(pos: ba.frozenbitarray) -> str:
        combined_mask = (
            sort_dimensions | unresolved_dimensions
            if unresolved_dimensions
            else sort_dimensions
        )
        return pos[combined_mask].to01()[::-1]

    positions.sort(key=_sort_key)
    # reverse so we can pop() from the back
    positions.reverse()
    return ChildGroupingTracker(sort_dimensions, unresolved_dimensions, positions)


# Backwards-compatible aliases
CoarseningTracker = ChildGroupingTracker


@copying_lru_cache()
def get_initial_coarsening_stack(
    current_parent_refinement: ba.frozenbitarray,
    dimensions_to_coarsen: ba.frozenbitarray,
    dimensions_cannot_coarsen: ba.frozenbitarray | None = None,
    linearization: Linearization = MortonOrderLinearization(),
) -> ChildGroupingTracker:
    """Backwards-compatible wrapper: dimensions_to_coarsen = sort_dimensions."""
    return get_initial_child_grouping(
        current_parent_refinement,
        dimensions_to_coarsen,
        dimensions_cannot_coarsen,
        linearization,
    )


def get_initial_coarsen_refine_stack(
    current_parent_refinement: ba.frozenbitarray,
    dimensions_to_coarsen: ba.frozenbitarray,
    dimensions_to_refine: ba.frozenbitarray,
    dimensions_cannot_coarsen: ba.frozenbitarray | None = None,
    linearization: Linearization = MortonOrderLinearization(),
) -> CoarseningTracker:
    assert (dimensions_to_coarsen & ~current_parent_refinement).count() == 0
    if dimensions_cannot_coarsen is not None:
        assert dimensions_cannot_coarsen.count() > 0
        assert (dimensions_cannot_coarsen & dimensions_to_refine).count() == 0
        assert (dimensions_cannot_coarsen & dimensions_to_coarsen).count() == 0
        assert (dimensions_cannot_coarsen & current_parent_refinement).count() == 0
    assert (dimensions_to_refine & current_parent_refinement).count() == 0

    later_iterated_dimensions = current_parent_refinement | dimensions_to_refine
    return get_initial_child_grouping(
        later_iterated_dimensions,
        dimensions_to_coarsen,
        dimensions_cannot_coarsen,
        linearization,
    )



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import warnings

from pathlib import Path
from typing import Sequence, Union, Mapping, Optional
import subprocess

from dyada.coordinates import (
    CoordinateInterval,
)
from dyada.descriptor import branch_generator, RefinementDescriptor
from dyada.discretization import (
    Discretization,
    discretization_to_location_stack_strings,
)
from dyada.drawing_util import (
    side_corners_generator,
    get_colors_byte,
    letter_counter,
)


def latex_write_and_compile(latex_string: str, filename: str) -> None:
    dirname = Path.cwd()
    with open(dirname / filename, "w") as f:
        f.write(latex_string)

    # if in environment, run pdflatex to generate pdf
    # this needs `latexmk`, `pdflatex`, and `tikz` to be installed, e.g.
    # through `apt install texlive-latex-extra latexmk`
    try:
        subprocess.check_output(
            ["latexmk", "-interaction=nonstopmode", "-pdf", filename],
            cwd=dirname,
        )
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        warnings.warn(f"Error while running latexmk on {filename}: {e}")
    try:  # try to clean up regardless of success
        subprocess.run(
            ["latexmk", "-c"],
            check=False,
            cwd=dirname,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass


def latex_add_color_defs(
    tikz_string: str, num_colors: int, colormap_name="CET_R3"
) -> str:
    for leaf, color in enumerate(get_colors_byte(num_colors, colormap_name)):
        color_str = "color_%d" % leaf
        tikz_string += "\\definecolor{%s}{RGB}{%d,%d,%d}\n" % (
            color_str,
            color[0],
            color[1],
            color[2],
        )
    return tikz_string


def plot_boxes_2d_tikz(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    filename: Optional[str] = None,
    connect_centers=False,
    **kwargs,
) -> None:
    tikz_string = R"""\documentclass{standalone}
% autogenerated with dyada : https://github.com/freifrauvonbleifrei/DyAda
\usepackage{xcolor}
\usepackage{relsize}
\usepackage{tikz}
\begin{document}
\begin{tikzpicture}
"""
    # add one white rectangle to the background
    tikz_string += "\\draw[very thin, white] (0,0) rectangle (1,1);\n"
    letter_counter_first = letter_counter(len(intervals))

    def tikz_rectangle(interval: CoordinateInterval, option_string="", label_string=""):
        lower = interval[0][projection]
        upper = interval[1][projection]
        rectangle_string = "\\draw[%s] (%f,%f) rectangle (%f,%f);\n" % (
            option_string,
            *lower,
            *upper,
        )
        middle = (lower + upper) / 2.0
        min_extent = min(upper - lower)  # type: ignore
        if label_string != "":
            if min_extent <= 0.125:
                label_string = "\\tiny \\relsize{-1} " + label_string
            elif min_extent <= 0.25:
                label_string = "\\tiny \\relsize{-0.5}" + label_string
            elif min_extent <= 0.5:
                label_string = "\\footnotesize " + label_string
        label_tex = f"\\node[inner sep=0] ({next(letter_counter_first)}) at ({middle[0]},{middle[1]}) {{{label_string}}};\n"

        return rectangle_string + label_tex

    if labels is None:

        def none_iter():
            while True:
                yield ""

        label_iter = none_iter()
    else:
        label_iter = iter(labels)
    tikz_string = latex_add_color_defs(tikz_string, len(intervals))
    for grid_idx, interval in enumerate(intervals):
        color_str = "color_%d" % grid_idx
        option_string = "very thin, gray, fill=%s,fill opacity=0.3" % (color_str)
        tikz_string += tikz_rectangle(interval, option_string, next(label_iter))

    if connect_centers:
        # connect the named nodes with lines
        letter_counter_again = letter_counter(len(intervals))
        connect_string = "\\draw[thick, densely dotted] "
        for node_name in letter_counter_again:
            connect_string += f"({node_name}) -- "
        tikz_string += connect_string[:-4] + ";\n"

    tikz_string += "\\end{tikzpicture}\n"
    tikz_string += "\\end{document}\n"
    if filename is None:
        filename = "tikz_rectangles"
    filename += ".tex"
    latex_write_and_compile(tikz_string, filename)


# inspired by @griegler : https://github.com/griegler/octnet/issues/28
def plot_boxes_3d_tikz(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    wireframe: bool = False,
    filename: Optional[str] = None,
    **kwargs,
) -> None:

    def tikz_cuboid(
        interval: CoordinateInterval, option_string="", label_string=""
    ) -> str:
        tikz_string = ""
        line_string = "\\draw[%s] (%f,%f,%f) -- (%f,%f,%f) -- (%f,%f,%f) -- (%f,%f,%f) -- cycle;\n"
        lower = interval[0][projection]
        upper = interval[1][projection]
        for side_corners in side_corners_generator(lower, upper):
            tikz_string += line_string % (
                option_string,
                *side_corners[0],
                *side_corners[1],
                *side_corners[2],
                *side_corners[3],
            )
        middle = (lower + upper) / 2.0
        min_extent = min(upper - lower)  # type: ignore
        if min_extent < 0.125:
            label_string = "\\tiny \\relsize{-1} " + label_string
        elif min_extent < 0.25:
            label_string = "\\tiny \\relsize{-0.5}" + label_string
        elif min_extent < 0.5:
            label_string = "\\footnotesize " + label_string
        tikz_string += (
            f"\\node at ({middle[0]},{middle[1]},{middle[2]}) {{{label_string}}};\n"
        )
        return tikz_string

    def tikz_grid(intervals, wireframe: bool):
        tikz_string = R"""\documentclass{standalone}
% autogenerated with dyada : https://github.com/freifrauvonbleifrei/DyAda
% inspired by @griegler : https://github.com/griegler/octnet/issues/28
\usepackage{xcolor}
\usepackage{relsize}
\usepackage{tikz,tikz-3dplot}
\begin{document}
\tdplotsetmaincoords{50}{130}
\begin{tikzpicture}[scale=1.0, tdplot_main_coords]"""
        if labels is None:

            def none_iter():
                while True:
                    yield ""

            label_iter = none_iter()
        else:
            label_iter = iter(labels)
        tikz_string = latex_add_color_defs(tikz_string, len(intervals))
        for grid_idx, interval in enumerate(intervals):
            color_str = "color_%d" % grid_idx
            if wireframe:
                option_string = "very thin, %s" % (color_str)
            else:
                option_string = "very thin, gray, fill=%s,fill opacity=0.3" % (
                    color_str
                )
            tikz_string += tikz_cuboid(interval, option_string, next(label_iter))

        tikz_string += "\\end{tikzpicture}\n"
        tikz_string += "\\end{document}\n"
        return tikz_string

    latex_string = tikz_grid(intervals, wireframe)
    if filename is None:
        filename = "tikz_cuboids"
    if wireframe:
        filename += "_wireframe.tex"
    else:
        filename += "_solid.tex"
    latex_write_and_compile(latex_string, filename)


def plot_tree_tikz(
    refinement_descriptor: RefinementDescriptor, labels=None, filename="omnitree"
):
    tikz_string = R"""\documentclass{standalone}
% autogenerated with dyada : https://github.com/freifrauvonbleifrei/DyAda
\usepackage{forest}
\begin{document}
% cf. https://tex.stackexchange.com/questions/332300/draw-lines-on-top-of-tikz-forest
\forestset{%
    declare keylist register={through},
    through={},
    tracing tree/.style={%
    delay={%
        for #1={%
        if phantom={}{through+/.option=name},
        }
    },
    before drawing tree={%
        tikz+/.wrap pgfmath arg={%
        \foreach \i [count=\j, remember=\i as \k] in {##1} \ifnum\j>1 \draw [densely dashed, ->] (\k.west) -- (\i.west)\fi;
        }{(through)}
    },
    }
}
\tikzset{every label/.style={xshift=-2ex,yshift=-2ex , font=\footnotesize, text=red}}
"""
    n_leaves = refinement_descriptor.get_num_boxes()
    tikz_string = latex_add_color_defs(tikz_string, n_leaves)

    assert labels is None or len(labels) == len(refinement_descriptor)

    tikz_string += "\\begin{forest}\n     tracing tree=tree,\n"
    d_zeros = "0" * refinement_descriptor.get_num_dimensions()
    leaf_string = (
        "[" + d_zeros + ",circle,draw,fill=color_%d,fill opacity=0.4,label={%s}]\n"
    )
    last_indent = 0
    tab = "    "
    num_box = 0
    num_patch = 0
    try:
        for current_branch, refinement in branch_generator(refinement_descriptor):
            current_indent = len(current_branch)
            while current_indent < last_indent:
                tikz_string += tab * current_indent + "]\n"
                last_indent -= 1
            if labels is not None:
                label_string = labels[num_patch]
            else:
                label_string = str(num_patch)
            if refinement.count() == 0:
                tikz_string += tab * current_indent + leaf_string % (
                    num_box,
                    label_string,
                )
                num_box += 1
            else:
                tikz_string += (
                    tab * current_indent
                    + "["
                    + refinement.to01()
                    + ",label={"
                    + label_string
                    + "}\n"
                )
            last_indent = current_indent
            num_patch += 1
    except IndexError:
        # if this happens, the descriptor is not valid but we can plot part of it
        warnings.warn("Descriptor is not valid, plotting as much as possible.")
        pass

    tikz_string += "]" * (last_indent)
    tikz_string += "\n\\end{forest}\n\\end{document}"

    if not filename.endswith(".tex"):
        filename += ".tex"
    latex_write_and_compile(tikz_string, filename)


def plot_descriptor_tikz(
    refinement_descriptor: RefinementDescriptor, filename="descriptor"
):
    tikz_string = R"""\documentclass{standalone}
% autogenerated with dyada : https://github.com/freifrauvonbleifrei/DyAda
\usepackage{tikz}
\usetikzlibrary{matrix}
\begin{document}"""
    num_colors = refinement_descriptor.get_num_boxes()
    tikz_string = latex_add_color_defs(tikz_string, num_colors)
    tikz_string += R"""\begin{tikzpicture}[every node/.style={draw,align=center,text height=2ex,minimum width=2ex, inner sep=0.2ex, fill opacity=0.4, text opacity=1}]
    \matrix [draw=none, matrix of nodes,nodes in empty cells]
    {"""
    tab = "    "
    box_counter = 0
    for refinement in refinement_descriptor:
        tikz_string += tab
        if refinement.count() == 0:
            tikz_string += f"|[fill=color_{box_counter}]| "
            box_counter += 1
        tikz_string += refinement.to01() + "&\n"

    # remove last newline and ampersand
    tikz_string = tikz_string[:-2]
    tikz_string += R"""\\
    };   
\end{tikzpicture}
\end{document}
"""
    if not filename.endswith(".tex"):
        filename += ".tex"
    latex_write_and_compile(tikz_string, filename)


def plot_location_stack_tikz(discretization: Discretization, filename="location_stack"):
    descriptor = discretization.descriptor
    location_stack = discretization_to_location_stack_strings(discretization)

    tikz_string = R"""\documentclass[tikz]{standalone}
% autogenerated with dyada : https://github.com/freifrauvonbleifrei/DyAda
\usepackage{tikz}
\usetikzlibrary{matrix}
\begin{document}"""
    num_colors = descriptor.get_num_boxes()
    tikz_string = latex_add_color_defs(tikz_string, num_colors)
    tikz_string += R"""\begin{tikzpicture}[
       every node/.style={align=left,anchor=west,
       text height=2ex,minimum width=2ex,
       inner sep=0.2ex,
       fill opacity=0.4, text opacity=1}]
    \matrix [
       draw=none, matrix of nodes] (n)
    {"""
    tab = "    "
    box_counter = 0
    for location_code_tuple, refinement in zip(location_stack, descriptor):
        tikz_string += tab
        for dimension in range(descriptor.get_num_dimensions()):
            location_code_string = "\\texttt{" + location_code_tuple[dimension] + "}"
            if dimension < descriptor.get_num_dimensions() - 1:
                location_code_string = location_code_string + "&"
            location_code_string = location_code_string.replace(
                "∩", "\\scalebox{0.79}{$\\cap$}"
            )
            location_code_string = location_code_string.replace(
                "λ", "\\scalebox{0.85}{$\\lambda$}"
            )

            if refinement.count() == 0 and len(location_code_tuple[dimension]) > 0:
                tikz_string += f"|[fill=color_{box_counter}]| "
            tikz_string += location_code_string
        tikz_string += "\\\\ \n"
        if refinement.count() == 0:
            box_counter += 1
    tikz_string = tikz_string[:-2]
    tikz_string += "\n};\n"
    tikz_string += "\\draw "
    for column in range(descriptor.get_num_dimensions()):
        tikz_string += f"(n-1-{column+1}.north west) -- (n-{len(location_stack)}-{column+1}.south west)"
    tikz_string += ";"
    tikz_string += R"""
\end{tikzpicture}
\end{document}
"""
    if not filename.endswith(".tex"):
        filename += ".tex"
    latex_write_and_compile(tikz_string, filename)



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

# %% [markdown]
# You can start with a regular `RefinementDescriptor`:

# %%
# code_embedder:descriptor start
import bitarray as ba
import dyada
from random import randint

# %%
descriptor = dyada.RefinementDescriptor(2, [2, 1])
# dyada.plot_tree_tikz(descriptor, filename="simple_tree")
num_dimensions = descriptor.get_num_dimensions()
print(descriptor)
# code_embedder:descriptor end

# %% [markdown]
# ```python
# RefinementDescriptor('11 01 00 00 ...0 00 01 00 00')
# ```


# %% [markdown]
# This one has four rectangles in the first dimension and two on the second, because
# the level `[2, 1]` is passed as base-2 exponents.
# If you uncomment the line with `plot_tree_tikz` and you have `latexmk` and some
# LaTeX tikz packages installed, the script will generate a `simple_tree.pdf` in the
# same folder.

# %% [markdown]
# You can use the descriptor and `MortonOrderLinearization` to build a `Discretization`:

# %%
# code_embedder:discretization start
discretization = dyada.Discretization(dyada.MortonOrderLinearization(), descriptor)
print("initial discretization:")
print(discretization)
# code_embedder:discretization end

# %% [markdown]
# ```python
# initial discretization:
# _________
# |_|_|_|_|
# |_|_|_|_|
# ```

# %% [markdown]
# If you want to refine a single rectangle at once, you can use `apply_single_refinement`:

# %%
# code_embedder:zero start
new_discretization, index_mapping = dyada.apply_single_refinement(
    discretization, 0, track_mapping="boxes"
)
print("after refining box 0:")
print(new_discretization)
# code_embedder:zero end

# %% [markdown]
# ```python
# after refining box 0:
# _________________
# |   |   |   |   |
# |___|___|___|___|
# |_|_|   |   |   |
# |_|_|___|___|___|
# ```


# %% [markdown]
# Of course, you can also refine only in a subset of the dimensions:

# %%
# code_embedder:random start
# select random index and refinement
random_index = randint(0, new_discretization.descriptor.get_num_boxes() - 1)
random_refinement = ba.bitarray("00")
while random_refinement.count() == 0:
    random_refinement = ba.bitarray(
        "".join(str(randint(0, 1)) for _ in range(num_dimensions))
    )
new_discretization, index_mapping = dyada.apply_single_refinement(
    new_discretization, random_index, random_refinement, track_mapping="boxes"
)
print("after refining random box:")
print(new_discretization)
# code_embedder:random end

# %% [markdown]
# ```python
# after refining random box:
# _________________________________
# |       |       |       |       |
# |_______|_______|_______|_______|
# |___|_|_|   |   |       |       |
# |___|___|___|___|_______|_______|
# ```


# %% [markdown]
# You can keep running the above and watch your discretization become finer and finer!

# %% [markdown]
# To refine many rectangles at once, you can collect the refinements
# as `PlannedAdaptiveRefinement` object:

# %%
# code_embedder:planned start
refining = dyada.PlannedAdaptiveRefinement(discretization)
refining.plan_refinement(0, ba.bitarray("11"))
refining.plan_refinement(1, ba.bitarray("01"))
new_discretization, index_mapping = refining.apply_refinements(track_mapping="boxes")
# dyada.plot_all_boxes_2d(new_discretization, backend="matplotlib", labels="boxes")
print("after applying planned refinements:")
print(new_discretization)
# code_embedder:planned end

# %% [markdown]
# ```python
# after applying planned refinements:
# _________________
# |   |   |   |   |
# |___|___|___|___|
# |_|_| | |   |   |
# |_|_|_|_|___|___|
# ```


# %% [markdown]
# If you uncomment the `plot_all_boxes_2d`, it will show you the discretization
# as matplotlib. Other backends are `tikz`, `ascii`(only 2d), and `opengl` (only 3d).

# %%
# TODO when adding to this file, update code_embedder references in readme



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import bitarray as ba
import numpy as np
from os.path import abspath

from dyada.coordinates import location_code_from_level_index

from dyada.descriptor import (
    RefinementDescriptor,
    validate_descriptor,
)
from dyada.discretization import (
    Discretization,
    SliceDictInDimension,
    discretization_to_location_stack_strings,
)
from dyada.linearization import MortonOrderLinearization


def test_get_level_index():
    r = Discretization(MortonOrderLinearization(), RefinementDescriptor(2, [1, 2]))

    # with indices that consider the parents
    level, index = r.get_level_index(0, False)
    assert np.array_equal(level, [0, 0]) and np.array_equal(index, [0, 0])
    level, index = r.get_level_index(1, False)
    assert np.array_equal(level, [1, 1]) and np.array_equal(index, [0, 0])
    level, index = r.get_level_index(2, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 0])
    level, index = r.get_level_index(3, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 1])
    level, index = r.get_level_index(4, False)
    assert np.array_equal(level, [1, 1]) and np.array_equal(index, [1, 0])
    level, index = r.get_level_index(5, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 0])
    level, index = r.get_level_index(6, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 1])
    level, index = r.get_level_index(7, False)
    assert np.array_equal(level, [1, 1]) and np.array_equal(index, [0, 1])
    level, index = r.get_level_index(8, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 2])
    level, index = r.get_level_index(9, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 3])
    level, index = r.get_level_index(10, False)
    assert np.array_equal(level, [1, 1]) and np.array_equal(index, [1, 1])
    level, index = r.get_level_index(11, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 2])
    level, index = r.get_level_index(12, False)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 3])
    with pytest.raises(IndexError):
        r.get_level_index(13, False)

    # with indices that consider only leaves / boxes
    level, index = r.get_level_index(0, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 0])
    level, index = r.get_level_index(1, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 1])
    level, index = r.get_level_index(2, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 0])
    level, index = r.get_level_index(3, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 1])
    level, index = r.get_level_index(4, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 2])
    level, index = r.get_level_index(5, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [0, 3])
    level, index = r.get_level_index(6, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 2])
    level, index = r.get_level_index(7, True)
    assert np.array_equal(level, [1, 2]) and np.array_equal(index, [1, 3])
    with pytest.raises(IndexError):
        r.get_level_index(8, True)


def test_get_all_boxes_level_indices():
    # same example as test_get_level_index
    r = Discretization(MortonOrderLinearization(), RefinementDescriptor(2, [1, 2]))
    expected_indices = ([0, 0], [0, 1], [1, 0], [1, 1], [0, 2], [0, 3], [1, 2], [1, 3])
    for i, level_index in enumerate(r.get_all_boxes_level_indices()):
        assert np.array_equal(level_index.d_level, np.asarray([1, 2]))
        assert np.array_equal(level_index.d_index, np.asarray(expected_indices[i]))


def test_get_box_from_coordinate():
    r = Discretization(MortonOrderLinearization(), RefinementDescriptor(2, [1, 2]))
    # check the midpoints of each box
    assert r.get_containing_box(np.array([0.25, 0.125])) == 0
    assert r.get_containing_box(np.array([0.25, 0.375])) == 1
    assert r.get_containing_box(np.array([0.75, 0.125])) == 2
    assert r.get_containing_box(np.array([0.75, 0.375])) == 3
    assert r.get_containing_box(np.array([0.25, 0.625])) == 4
    assert r.get_containing_box(np.array([0.25, 0.875])) == 5
    assert r.get_containing_box(np.array([0.75, 0.625])) == 6
    assert r.get_containing_box(np.array([0.75, 0.875])) == 7

    # now along the edges
    assert r.get_containing_box(np.array([0.0, 0.0])) == 0
    assert r.get_containing_box(np.array([0.0, 0.25])) == (0, 1)
    assert r.get_containing_box(np.array([0.0, 0.5])) == (1, 4)
    assert r.get_containing_box(np.array([0.0, 0.75])) == (4, 5)
    assert r.get_containing_box(np.array([0.0, 1.0])) == 5
    assert r.get_containing_box(np.array([0.25, 0.0])) == 0
    assert r.get_containing_box(np.array([0.25, 0.25])) == (0, 1)
    assert r.get_containing_box(np.array([0.25, 0.5])) == (1, 4)
    assert r.get_containing_box(np.array([0.25, 0.75])) == (4, 5)
    assert r.get_containing_box(np.array([0.25, 1.0])) == 5
    assert r.get_containing_box(np.array([0.5, 0.0])) == (0, 2)
    assert r.get_containing_box(np.array([0.5, 0.25])) == (0, 1, 2, 3)
    assert r.get_containing_box(np.array([0.5, 0.5])) == (1, 3, 4, 6)
    assert r.get_containing_box(np.array([0.5, 0.75])) == (4, 5, 6, 7)
    assert r.get_containing_box(np.array([0.5, 1.0])) == (5, 7)
    assert r.get_containing_box(np.array([0.75, 0.0])) == 2
    assert r.get_containing_box(np.array([0.75, 0.25])) == (2, 3)
    assert r.get_containing_box(np.array([0.75, 0.5])) == (3, 6)
    assert r.get_containing_box(np.array([0.75, 0.75])) == (6, 7)
    assert r.get_containing_box(np.array([0.75, 1.0])) == 7
    assert r.get_containing_box(np.array([1.0, 0.0])) == 2
    assert r.get_containing_box(np.array([1.0, 0.25])) == (2, 3)
    assert r.get_containing_box(np.array([1.0, 0.5])) == (3, 6)
    assert r.get_containing_box(np.array([1.0, 0.75])) == (6, 7)
    assert r.get_containing_box(np.array([1.0, 1.0])) == 7

    # outside the domain
    with pytest.raises(ValueError):
        r.get_containing_box(np.array([0.0, 1.5]))
    with pytest.raises(ValueError):
        r.get_containing_box(np.array([1.5, 0.0]))
    with pytest.raises(ValueError):
        r.get_containing_box(np.array([1.5, 1.5]))


def test_get_index_from_location_code_from_level_index_2d():
    descriptor = RefinementDescriptor.from_binary(  # same as test_refine_2_3
        2, ba.bitarray("11 00 10 01 00 00 01 00 00 00 10 00 01 00 00")
    )
    validate_descriptor(descriptor)
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    # test all indices
    for i in range(len(descriptor)):
        level_index = discretization.get_level_index(i, is_box_index=False)
        location_code = location_code_from_level_index(level_index)
        index = discretization.get_index_from_location_code(
            location_code, get_box=False
        )
        assert index == i  # expect exact match
        # now extend one location code by one level
        extended_location_code = (location_code[0] + ba.bitarray("1"), location_code[1])
        try:
            index = discretization.get_index_from_location_code(
                extended_location_code, get_box=False
            )
            assert index > i  # if match, it has to be a child
        except Discretization.NoExactMatchError as e:
            assert e.closest_match == i

        other_extended_location_code = (
            location_code[0],
            location_code[1] + ba.bitarray("1"),
        )
        try:
            index = discretization.get_index_from_location_code(
                other_extended_location_code, get_box=False
            )
            assert index > i  # if match, it has to be a child
        except Discretization.NoExactMatchError as e:
            assert e.closest_match == i


def helper_mapping_as_box_mapping(
    old_descriptor: RefinementDescriptor, new_descriptor: RefinementDescriptor, mapping
):
    """
    Helper function to assert that a mapping can be converted to a box mapping.
    """
    box_keys = list()
    box_values = list()
    for key, value in mapping.items():
        if not old_descriptor.is_box(key):
            continue
        box_keys.append(old_descriptor.to_box_index(key))
        assert new_descriptor.is_box(value)
        box_values.append(new_descriptor.to_box_index(value))

    assert len(set(box_keys)) == len(box_keys)
    assert len(set(box_values)) == len(box_values)

    box_mapping = {
        old_descriptor.to_box_index(k): new_descriptor.to_box_index(v)
        for k, v in mapping.items()
        if old_descriptor.is_box(k)
    }
    assert set(box_mapping.keys()) == set(box_keys)


def test_slice_discretization_3d():
    # same 3d discretization as in test_plot_boxes_3d_from_descriptor
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "101 000 001 000 000 010 100 000 000 000"
            "101 000 000 010 000 101 000 000 000 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)

    # # plot
    # plot_all_boxes_3d(
    #     discretization,
    #     filename="test_slice_discretization_3d",
    #     labels="patches",
    #     title="test_slice_discretization_3d",
    #     backend="tikz",
    # )

    for z_i in range(0, 20):
        z = z_i / 19
        discretization_xy, mapping_xy = discretization.slice([None, None, z])
        assert validate_descriptor(discretization_xy.descriptor)
        helper_mapping_as_box_mapping(
            descriptor, discretization_xy.descriptor, mapping_xy
        )
        # plot_all_boxes_2d(
        #     discretization_xy,
        #     filename="test_slice_discretization_3d_z" + str(z_i),
        #     labels="patches",
        #     title="test_slice_discretization_3d_z" + str(z_i),
        #     backend="tikz",
        # )

        for x_i in range(0, 20):
            x = x_i / 19
            discretization_y, mapping_y = discretization_xy.slice([x, None])
            assert validate_descriptor(discretization_y.descriptor)
            helper_mapping_as_box_mapping(
                discretization_xy.descriptor,
                discretization_y.descriptor,
                mapping_y,
            )
            mapping_z_to_x_to_y = {}
            for old_y, new_y in mapping_y.items():
                # invert the xy mapping, find old_y
                for old_xy, new_xy in mapping_xy.items():
                    if new_xy == old_y:
                        mapping_z_to_x_to_y[old_xy] = new_y
                        break
            # check if slicing in two steps is the same as slicing in one step
            discretization_y_at_once, mapping_y_at_once = discretization.slice(
                [x, None, z]
            )
            helper_mapping_as_box_mapping(
                discretization.descriptor,
                discretization_y_at_once.descriptor,
                mapping_y_at_once,
            )
            # assert that everything in mapping_z_to_x_to_y is in mapping_y
            assert all(
                [
                    mapping_y_at_once[old_y] == new_y
                    for old_y, new_y in mapping_z_to_x_to_y.items()
                ]
            )
            assert validate_descriptor(discretization_y_at_once.descriptor)
            assert discretization_y == discretization_y_at_once


def test_all_slices_3d():
    # same 3d discretization as in test_plot_boxes_3d_from_descriptor
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "101 000 001 000 000 010 100 000 000 000"
            "101 000 000 010 000 101 000 000 000 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)

    z_start = 0.0
    x_start = 0.0
    z_used = set()
    x_used = set()
    while z_start < 1.0 and x_start < 1.0:
        # slice at z_start and x_start
        discretization_xz, mapping_xz, levels_xz = discretization.slice(
            [x_start, None, z_start], get_level=True
        )
        z_used.add(z_start)
        x_used.add(x_start)
        assert validate_descriptor(discretization_xz.descriptor)
        x_start += 2.0 ** -levels_xz[0]
        z_start += 2.0 ** -levels_xz[2]

        helper_mapping_as_box_mapping(
            descriptor, discretization_xz.descriptor, mapping_xz
        )

    assert len(z_used) == 3
    assert len(x_used) == 3


def test_hash_discretization():
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "101 000 001 000 000 010 100 000 000 000"
            "101 000 000 010 000 101 000 000 000 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    other_discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert hash(discretization) == hash(other_discretization)
    different_descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "101 000 001 000 000 010 100 000 000 000 101 000 000 010 000 000 000"
        ),
    )
    different_discretization = Discretization(
        MortonOrderLinearization(), different_descriptor
    )
    assert hash(discretization) != hash(different_discretization)


def test_slice_dict_3d():
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "101 000 001 000 000 010 100 000 000 000"
            "101 000 000 010 000 101 000 000 000 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    slice_dict_z = SliceDictInDimension(discretization, 2, True)
    assert slice_dict_z.keys() == {0.0, 0.25, 0.5, 0.75, 0.875}
    for key, item in slice_dict_z.items():
        assert key in slice_dict_z.keys()
        assert isinstance(item[0], Discretization)
        assert item[0].descriptor.get_num_dimensions() == 2
        assert isinstance(item[1], dict)
    assert slice_dict_z[0.1] is slice_dict_z[0.0]
    assert slice_dict_z[0.4999] is slice_dict_z[0.25]
    assert slice_dict_z[1.0] is slice_dict_z[0.875]
    with pytest.raises(KeyError):
        slice_dict_z[1.00001]
    # try to store other data types as values
    for key in slice_dict_z.keys():
        slice_dict_z[key] = len(slice_dict_z[key][0])
    assert slice_dict_z[0.0] == 2
    assert slice_dict_z[0.25] == 2
    assert slice_dict_z[0.5] == 5
    assert slice_dict_z[0.75] == 7
    assert slice_dict_z[0.875] == 7


def test_stack():
    first_descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 00 00 10 01 00 00 00")
    )
    assert discretization_to_location_stack_strings(
        Discretization(MortonOrderLinearization(), first_descriptor)
    ) == [
        ("λ", ""),
        ("0", "λ"),
        ("0", "0"),
        ("0", "1"),
        ("1λ", ""),
        ("10", "λ"),
        ("10", "0"),
        ("10", "1"),
        ("11", ""),
    ]
    second_descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 10 01 00 00 01 00 00 00 10 00 01 00 00")
    )
    assert discretization_to_location_stack_strings(
        Discretization(MortonOrderLinearization(), second_descriptor)
    ) == [
        ("λ", "λ"),
        ("0", "0"),
        ("1λ", "0"),
        ("10", "0λ"),
        ("10", "00"),
        ("10", "01"),
        ("11", "0λ"),
        ("11", "00"),
        ("11", "01"),
        ("0", "1"),
        ("1λ", "1"),
        ("10", "1"),
        ("11", "1λ"),
        ("11", "10"),
        ("11", "11"),
    ]
    third_descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 11 00 00 00 00 00 10 00 01 00 00")
    )
    assert discretization_to_location_stack_strings(
        Discretization(MortonOrderLinearization(), third_descriptor)
    ) == [
        ("λ", "λ"),
        ("0", "0"),
        ("1λ", "0λ"),
        ("10", "00"),
        ("11", "00"),
        ("10", "01"),
        ("11", "01"),
        ("0", "1"),
        ("1λ", "1"),
        ("10", "1"),
        ("11", "1λ"),
        ("11", "10"),
        ("11", "11"),
    ]
    fourth_descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 00 00 10 00 00")
    )
    assert discretization_to_location_stack_strings(
        Discretization(MortonOrderLinearization(), fourth_descriptor)
    ) == [
        ("λ", ""),
        ("0", "λ"),
        ("0", "0"),
        ("0", "1"),
        ("1λ", ""),
        ("10", ""),
        ("11", ""),
    ]


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import warnings

try:
    from matplotlib.colors import to_rgb
except ImportError:
    warnings.warn("matplotlib not found, some plotting functions will not work")
try:
    import OpenGL.GL as gl  # type: ignore
    import OpenGL.GLU as glu  # type: ignore
    import OpenGL.GLUT as glut  # type: ignore
    from PIL import Image, ImageOps
except ImportError:
    warnings.warn("pyopengl not found, some plotting functions will not work")

from io import StringIO
from itertools import pairwise
from typing import Sequence, Union, Mapping, Optional

from dyada.coordinates import (
    CoordinateInterval,
)
from dyada.drawing_util import (
    cuboid_from_interval,
    side_corners_generator,
    get_colors,
)
from dyada.structure import depends_on_optional


@depends_on_optional("OpenGL")
def draw_cuboid_opengl(
    interval: CoordinateInterval,
    projection: Sequence[int],
    wireframe: bool = False,
    alpha: float = 0.1,
    linewidth: float = 1.0,
    color=(0.5, 0.5, 0.5),
):
    color = to_rgb(color)

    if wireframe:
        alpha_faces = 0.0
        alpha_lines = alpha
        color_faces = (0.0, 0.0, 0.0)
        color_lines = color
    else:
        alpha_faces = alpha
        alpha_lines = 1.0
        color_faces = color
        color_lines = (0.0, 0.0, 0.0)

    lower = interval[0][projection]
    upper = interval[1][projection]

    gl.glLineWidth(linewidth)
    gl.glBegin(gl.GL_LINES)
    gl.glColor4fv((*color_lines, alpha_lines))
    for side_corners in side_corners_generator(lower, upper):
        for side in pairwise(side_corners):
            gl.glVertex3fv(side[0])
            gl.glVertex3fv(side[1])
    gl.glEnd()

    gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)  # try to avoid overdrawing
    gl.glPolygonOffset(1.0, 1.0)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor4fv((*color_faces, alpha_faces))
    for side_corners in side_corners_generator(lower, upper):
        for corner in side_corners:
            gl.glVertex3fv(corner)
    gl.glEnd()
    gl.glDisable(gl.GL_POLYGON_OFFSET_FILL)


@depends_on_optional("OpenGL")
def gl_save_file(filename: str, width=1024, height=1024) -> None:
    gl.glFlush()
    gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
    data = gl.glReadPixels(0, 0, width, height, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE)
    image = Image.frombytes("RGBA", (width, height), data)
    image = ImageOps.flip(image)
    if not filename.endswith(".png"):
        filename += ".png"
    image.save(filename, "PNG")


@depends_on_optional("OpenGL")
def plot_boxes_3d_pyopengl(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    labels: Optional[Sequence[str]],
    projection: Sequence[int],
    wireframe: bool = False,
    filename: str = "omnitree",
    width: int = 1024,
    height: int = 1024,
    alpha: float = 0.1,
    linewidth: float = 1.0,
    **kwargs,
) -> None:
    if labels is not None:
        warnings.warn("Labels are currently not used in 3D plots w/ pyopengl")

    colors = kwargs.pop("colors", get_colors(len(intervals)))
    if isinstance(colors, str):
        colors = [colors] * len(intervals)

    def init_glu():
        glut.glutInit()
        glut.glutInitDisplayMode(glut.GLUT_DOUBLE | glut.GLUT_RGB)
        glut.glutInitWindowSize(width, height)
        glut.glutCreateWindow(b"Dyada OpenGL")
        glut.glutHideWindow()
        gl.glClearColor(1.0, 1.0, 1.0, 1.0)
        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        glu.gluPerspective(
            35.0,
            float(width) / height,
            0.1,
            10.0,
        )
        gl.glMatrixMode(gl.GL_MODELVIEW)
        # move camera back to see the unit cube
        glu.gluLookAt(
            *(-1.5, 1.5, 3.0),
            *(0.5, 0.5, 0.5),
            *(0, 1, 0),
        )
        # for opacity
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        # draw background
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

    init_glu()

    for interval, color in zip(intervals, colors):
        draw_cuboid_opengl(
            interval,
            projection,
            wireframe=wireframe,
            alpha=alpha,
            linewidth=linewidth,
            color=color,
        )

    if filename is not None:
        gl_save_file(filename, width, height)


def add_cuboid_to_buffer(
    buffer: StringIO,
    vertex_offset: int,
    interval: CoordinateInterval,
    projection: Sequence[int],
    wireframe: bool = False,
):
    verts, faces, edges = cuboid_from_interval(interval, projection)
    # Write vertices
    for v in verts:
        buffer.write(f"v {v[0]} {v[1]} {v[2]}\n")

    if wireframe:
        # Write edges
        for edge in edges:
            a, b = (vertex_offset + i for i in edge)
            buffer.write(f"l {a} {b}\n")
    else:
        # Write faces
        for quad in faces:
            a, b, c, d = (vertex_offset + i for i in quad)
            buffer.write(f"f {a} {b} {c} {d}\n")

    vertex_offset += len(verts)  # Move offset for next cuboid
    return buffer, vertex_offset


def write_obj_file(buffer: StringIO, filename: str):
    if not filename.endswith(".obj"):
        filename += ".obj"
    with open(filename, "w") as f:
        return f.write(buffer.getvalue())


def export_boxes_3d_to_obj(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    projection: Sequence[int],
    wireframe: bool = False,
    filename: str = "omnitree",
    **kwargs,
):
    buffer = StringIO()
    # .obj is 1-indexed
    vertex_offset = 1
    for interval in intervals:
        buffer, vertex_offset = add_cuboid_to_buffer(
            buffer, vertex_offset, interval, projection, wireframe
        )

    if filename is None:
        return buffer, vertex_offset
    return write_obj_file(buffer, filename)



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import bitarray as ba
from os.path import abspath

from dyada.descriptor import RefinementDescriptor, validate_descriptor
from dyada.refinement import (
    Discretization,
    PlannedAdaptiveRefinement,
    apply_single_refinement,
)
from dyada.linearization import MortonOrderLinearization

from test_refinement import helper_check_mapping


def test_refine_3d_1():
    prependable_string = "110001000000001000000001000000"
    for round_number in range(4):
        descriptor = RefinementDescriptor.from_binary(
            3,
            ba.bitarray(
                prependable_string * round_number
                + "110001000000001010000000000001010000000000000"
            ),
        )
        validate_descriptor(descriptor)
        num_boxes_before = descriptor.get_num_boxes()
        discretization = Discretization(MortonOrderLinearization(), descriptor)

        new_discretization, index_mapping = apply_single_refinement(
            discretization, len(discretization) - 1, ba.bitarray("001")
        )

        new_descriptor = new_discretization.descriptor
        validate_descriptor(new_descriptor)
        assert new_descriptor.get_num_boxes() == num_boxes_before + 1
        helper_check_mapping(index_mapping, discretization, new_discretization)
        # make sure only the last box maps to two new boxes
        for b in range(descriptor.get_num_boxes()):
            assert len(index_mapping[b]) == 1 or (
                b == (len(discretization) - 1) and len(index_mapping[b]) == 2
            )


def test_refine_3d_2():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(3, ba.bitarray("001 000 100 000 000")),
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "110")
    p.plan_refinement(1, "011")
    p.plan_refinement(2, "011")
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2, 3, 4},
        2: {0, 5, 8, 11, 14},
        3: {0, 5, 6, 7, 11, 12, 13},
        4: {0, 8, 9, 10, 14, 15, 16},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]
    helper_check_mapping(patch_mapping, discretization, new_discretization, False)


def test_refine_3d_3():
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("010 000 101 100 000 000 100 000 000 100 000 000 100 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "101")
    p.plan_refinement(1, "100")
    p.plan_refinement(2, "100")
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2, 13, 14},
        2: {0, 3, 10, 15, 18},
        3: {3, 4, 7},
        4: {4, 5, 6},
        5: {7, 8, 9},
        6: {10},
        7: {11},
        8: {12},
        9: {15},
        10: {16},
        11: {17},
        12: {18},
        13: {19},
        14: {20},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]
    helper_check_mapping(patch_mapping, discretization, new_discretization, False)


def test_refine_3d_4():
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("010 000 001 000 100 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "101")
    p.plan_refinement(1, "100")
    p.plan_refinement(2, "111")
    p.plan_refinement(3, "111")
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2, 5, 6},
        2: {0, 3, 4, 7, 16},
        3: {0, 3, 4},
        4: {0, 7, 16},
        5: {7, 8, 9, 10, 11, 12, 13, 14, 15},
        6: {16, 17, 18, 19, 20, 21, 22, 23, 24},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]
    helper_check_mapping(patch_mapping, discretization, new_discretization, False)


def test_refine_3d_5():
    # almost same beginning as 3d_4
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("001 010 100 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(2, "100")
    p.plan_refinement(3, "110")
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    expected_patch_mapping = {
        0: {0},
        1: {0, 3, 4},
        2: {0},
        3: {1},
        4: {2},
        5: {0, 3, 4},
        6: {0, 5, 6, 7, 8},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]
    helper_check_mapping(patch_mapping, discretization, new_discretization, False)


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from queue import PriorityQueue


import dyada.coordinates
import dyada.descriptor
import dyada.discretization
import dyada.drawing
import dyada.linearization
import dyada.refinement


def test_partitioning_queued():
    # random input tensor
    random_length = np.random.randint(10, 100)
    # values in the tensor correspond to "importance" of the data
    tensor = np.random.rand(240, 160, random_length)
    num_dimensions = tensor.ndim
    possible_refinements = list(dyada.linearization.single_bit_set_gen(num_dimensions))
    num_desired_partitions = 100
    minimum_partition_size = [16, 16, 1]

    def get_dimensionwise_importance(subtensor):
        # to assign the refinement importance per partition and dimension,
        # get the difference quotients for each dimension and add them to the absolute values
        # (can use other dimension-sensitive filters here,
        # can even be different for each dimension) #TODO
        importances = []
        for d in range(num_dimensions):
            difference_quotient = np.diff(subtensor, axis=d)
            # normalize the accumulated quotient by the size of the difference quotient array
            difference_quotient_normalized = np.sum(difference_quotient) / (
                difference_quotient.size * subtensor.size
            )
            importances.append(
                np.sum(np.abs(subtensor)) + difference_quotient_normalized
            )
        return importances

    # represent whole tensor domain by discretization
    discretization = dyada.discretization.Discretization(
        dyada.linearization.MortonOrderLinearization(),
        dyada.descriptor.RefinementDescriptor(num_dimensions, 0),
    )
    full_domain = dyada.coordinates.interval_from_sequences(
        [0] * num_dimensions, tensor.shape
    )
    while len(discretization) < num_desired_partitions:
        priority_queue = PriorityQueue()
        for i in range(len(discretization)):
            # what area does the i-th partition represent?
            interval = dyada.discretization.coordinates_from_box_index(
                discretization, i, full_domain
            )
            # use exactly the voxels, slightly more if necessary
            sub_tensor = tensor[
                tuple(
                    slice(
                        int(np.floor(interval.lower_bound[d])),
                        int(np.ceil(interval.upper_bound[d])),
                    )
                    for d in range(num_dimensions)
                )
            ]
            importance = get_dimensionwise_importance(sub_tensor)
            for d in range(num_dimensions):
                # skip if the partition would be too small
                if sub_tensor.shape[d] < minimum_partition_size[d] * 2:
                    continue
                priority_queue.put((-importance[d], (i, possible_refinements[d])))

        # partition the tensor into two parts based on the highest importance
        neg_importance, next_refinement = priority_queue.get()
        # can be sped up by re-mapping instead of recomputing all priorities
        discretization, index_mapping = dyada.refinement.apply_single_refinement(
            discretization, *next_refinement
        )

    # plot the final partitioning -- look for "patches_solid.pdf" in the current directory
    dyada.drawing.plot_all_boxes_3d(discretization, labels=None, filename="patches")

    # interpolation / integration of each partition to desired resolution... processing... other things...#TODO
    # dummy variant: take the mean in the last dimension, resample to 16 in the other dimensions
    partitions_resampled = []
    for i in range(len(discretization)):
        interval = dyada.discretization.coordinates_from_box_index(
            discretization, i, full_domain
        )
        sub_tensor = tensor[
            tuple(
                slice(
                    int(np.floor(interval.lower_bound[d])),
                    int(np.ceil(interval.upper_bound[d])),
                )
                for d in range(num_dimensions)
            )
        ]
        sub_tensor_averaged = np.mean(sub_tensor, axis=-1)
        x = np.arange(sub_tensor_averaged.shape[0])
        y = np.arange(sub_tensor_averaged.shape[1])
        try:
            from scipy.interpolate import RectBivariateSpline  # type: ignore

            # bicubic interpolation w/ scipy
            sub_tensor_spline = RectBivariateSpline(x, y, sub_tensor_averaged)
            x_new = np.linspace(0, sub_tensor_averaged.shape[0], 16)
            y_new = np.linspace(0, sub_tensor_averaged.shape[1], 16)
            sub_tensor_resampled = sub_tensor_spline(x_new, y_new)
        except ImportError:
            # fallback: subsampling
            sub_tensor_resampled = np.zeros((16, 16))
            for x_new in range(16):
                for y_new in range(16):
                    x_old = x_new * sub_tensor_averaged.shape[0] // 16
                    y_old = y_new * sub_tensor_averaged.shape[1] // 16
                    sub_tensor_resampled[x_new, y_new] = sub_tensor_averaged[
                        x_old, y_old
                    ]
        assert sub_tensor_resampled.shape == (16, 16)
        partitions_resampled.append(sub_tensor_resampled)

        # questions:
        # 1. do we need to do slicing (i.e., know which patches occur at a certain frame)?
        # 2. can/should neighbor information be included?



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
from collections import deque, Counter
from dataclasses import dataclass
from functools import cached_property
from itertools import islice
from re import findall
import numpy as np
import numpy.typing as npt
import operator
import reprlib
from typing import Iterator, Optional, Sequence, Union


# related to generalized (2^d-ary) ruler function, e.g. https://oeis.org/A115362
def get_regular_refined(added_level: Sequence[int]) -> ba.bitarray:
    num_dimensions = len(added_level)
    data = ba.bitarray(num_dimensions)

    # iterate in reverse from max(level) to 0...
    for level in reversed(range(max(added_level))):
        at_least_l = ba.bitarray([1 if i > level else 0 for i in added_level])
        # power of two by bitshift
        factor = 1 << at_least_l.count()
        # ...while duplicating the current data as new children
        data = at_least_l + data * factor

    return data


def get_num_children_from_refinement(refinement: ba.bitarray) -> int:
    num_ones = refinement.count()
    if num_ones == 0:
        return 0
    else:
        return 1 << num_ones


@dataclass
class LevelCounter:
    level_increment: ba.frozenbitarray
    count_to_go_up: int


class Branch(deque[LevelCounter]):
    """A branch points us to a location in a tree: it is a stack (implemented
    as deque) of LevelCounter "twigs" of level increments and a count that goes
    down as we progress in the number of siblings (this count is always
    initialized to the number of children of the node with the respective
    level_increment/refinement).
    Note that the refinement of the considered node itself is not part of
    the branch, there is only the parent refinement.
    """

    def __init__(self, num_dimensions_or_other_branch: Union[int, "Branch"]):
        if isinstance(num_dimensions_or_other_branch, int):
            super().__init__()
            num_dimensions = num_dimensions_or_other_branch
            dZeros = ba.frozenbitarray([0] * num_dimensions)
            self.append(LevelCounter(dZeros, 1))
        else:
            super().__init__(num_dimensions_or_other_branch)

    def advance_branch(self, check_depth: int = 0) -> None:
        """Advance the branch to the next sibling, in-place"""
        self[-1].count_to_go_up -= 1
        assert self[-1].count_to_go_up >= 0
        while self[-1].count_to_go_up == 0:
            self.pop()
            if len(self) <= check_depth:
                raise IndexError
            self[-1].count_to_go_up -= 1
            assert self[-1].count_to_go_up >= 0

    def grow_branch(self, level_increment: ba.frozenbitarray) -> None:
        """Go deeper in the branch hierarchy / add a twig"""
        # power of two by bitshift
        self.append(LevelCounter(level_increment, 1 << level_increment.count()))

    def __repr__(self) -> str:
        contents = ""
        for twig in self:
            contents += f"{twig.level_increment}({twig.count_to_go_up})-"
        return f"Branch({contents})"

    def to_history(self) -> tuple[list[int], list[ba.frozenbitarray]]:
        history_of_indices: list[int] = []
        history_of_level_increments: list[ba.frozenbitarray] = []
        for level_count in range(1, len(self)):
            current_refinement = self[level_count].level_increment
            # power of two by bitshift
            linear_index_at_level = (1 << current_refinement.count()) - self[
                level_count
            ].count_to_go_up
            history_of_level_increments.append(current_refinement)
            history_of_indices.append(linear_index_at_level)
        return history_of_indices, history_of_level_increments


class DyadaInvalidDescriptorError(Exception):
    pass


class RefinementDescriptor:
    """
    A RefinementDescriptor holds a bitarray that describes a refinement tree.
    The bitarray is a depth-first linearized 2^n tree, with the parents having
    the refined dimensions set to 1 and the leaves containing all 0s.
    It is the (preorder depth-first) linearized binary representation of omnitrees
    described in https://arxiv.org/abs/2508.06316 .
    """

    def __init__(self, num_dimensions: int, base_resolution_level=0):
        self._num_dimensions = num_dimensions
        if isinstance(base_resolution_level, int):
            base_resolution_level = [base_resolution_level] * self._num_dimensions
        assert len(base_resolution_level) == self._num_dimensions

        # establish the base resolution level
        self._data = get_regular_refined(base_resolution_level)

    def to_file(self, filename: str):
        filename_full = filename
        filename_full += "_{dim}d.bin".format(dim=self.get_num_dimensions())
        with open(filename_full, "wb") as file:
            # invert so that trailing zeroes are easier to remove
            inverted_data = ~self._data
            inverted_data.tofile(file)

    @staticmethod
    def from_file(filename: str):
        # extract dimensionality from filename
        num_dimensions_list = findall(r"_(\d+)d.bin", filename)
        num_dimensions = int(num_dimensions_list[0])
        inverted_data = ba.bitarray()
        with open(filename, "rb") as file:
            inverted_data.fromfile(file)
        # remove trailing zeroes
        num_trailing_zeros = 0
        while inverted_data[-num_trailing_zeros - 1] == 0:
            num_trailing_zeros += 1
        inverted_data = inverted_data[
            : None if num_trailing_zeros == 0 else -num_trailing_zeros
        ]
        return RefinementDescriptor.from_binary(num_dimensions, ~inverted_data)

    @staticmethod
    def from_binary(num_dimensions: int, binary: ba.bitarray) -> "RefinementDescriptor":
        if len(binary) % num_dimensions != 0:
            raise ValueError("Invalid binary input length")
        descriptor = RefinementDescriptor(num_dimensions)
        descriptor._data = binary
        try:
            validate_descriptor(descriptor)
        except DyadaInvalidDescriptorError as e:
            raise ValueError("Invalid binary input") from e
        return descriptor

    def __eq__(self, other):
        return (
            (self._num_dimensions == other._num_dimensions)
            and (len(self) == len(other))
            and (self._data == other._data)
        )

    def __hash__(self):
        return hash((self._num_dimensions, self._data.tobytes()))

    def __len__(self):
        """
        return the number of refinement descriptions, will be somewhere between get_num_boxes() and 2*get_num_boxes()
        """
        return len(self._data) // self._num_dimensions

    def get_num_dimensions(self):
        return self._num_dimensions

    @cached_property
    def d_zeros(self):
        return ba.frozenbitarray(self._num_dimensions)

    def get_num_boxes(self):
        # count number of d*(0) bit blocks
        return sum(x.count() == 0 for x in self)

    def get_data(self):
        return self._data

    def is_box(self, index: int):
        return self[index] == self.d_zeros

    def __repr__(self) -> str:
        return (
            f"RefinementDescriptor({reprlib.repr(' '.join([b.to01() for b in self]))})"
        )

    def __iter__(self, start: int = 0):
        if __debug__:  # slow and safe mode
            for i in range(start, len(self)):
                # same as yield ba.frozenbitarray(self[i])
                yield ba.frozenbitarray(
                    self.get_data()[
                        i * self._num_dimensions : (i + 1) * self._num_dimensions
                    ]
                )
            return
        # fast mode, won't work with the functions that use Counter directly
        j = self._num_dimensions * start
        for _ in range(start, len(self)):
            next_j = j + self._num_dimensions
            yield self.get_data()[j:next_j]
            j = next_j

    def __getitem__(self, index_or_slice):
        nd = self._num_dimensions
        if isinstance(index_or_slice, slice):
            assert index_or_slice.step == 1 or index_or_slice.step is None
            start = index_or_slice.start
            stop = index_or_slice.stop
            start = 0 if start is None else operator.index(start)
            stop = len(self) if stop is None else operator.index(stop)
            return self.get_data()[start * nd : stop * nd]
        else:  # it should be an index
            index_or_slice = operator.index(index_or_slice)
            return self.get_data()[index_or_slice * nd : (index_or_slice + 1) * nd]

    def is_pow2tree(self) -> bool:
        """Is this a quadtree / octree / general power-of-2 tree?"""
        c = Counter(self)
        return c.keys() == {
            self.d_zeros,
            ba.frozenbitarray("1" * self._num_dimensions),
        }

    def num_boxes_up_to(self, index: int) -> int:
        # count zeros up to index, zero-indexed
        # (Counter is also possible, but keeps us from performance opt. in __iter__)
        count = sum((x == self.d_zeros) for x in islice(self, index))
        return count

    def to_box_index(self, index: int) -> int:
        assert self.is_box(index)
        # count zeros up to index, zero-indexed
        return self.num_boxes_up_to(index)

    def _to_box_index_recursive(self, index: int) -> int:
        assert self.is_box(index)
        try:
            # iterate down until we find the prior box
            for i in range(index - 1, 0, -1):
                if self[i] == self.d_zeros:
                    return self._to_box_index_recursive(i) + 1
            # if we are at the first box, return 0
            return 0
        except RecursionError:
            return self.num_boxes_up_to(index)

    def get_maximum_level(self) -> npt.NDArray[np.int8]:
        max_level = np.max(list(self.level_iterator()), axis=0)
        return max_level

    def to_hierarchical_index(self, box_index: int) -> int:
        linear_index = 0
        # count down box index
        for i in self:
            if i == self.d_zeros:
                box_index -= 1
            if box_index < 0:
                break
            linear_index += 1
        assert self.is_box(linear_index)
        return linear_index

    def get_branch(
        self,
        index: int,
        is_box_index: bool = True,
        hint_previous_branch: tuple[int, Branch] | None = None,
    ) -> tuple[Branch, Iterator]:
        if index < 0 or index >= len(self):
            raise IndexError("Index out of range")

        box_counter = 0
        if hint_previous_branch is None:
            current_branch = Branch(self._num_dimensions)
            i = 0
            current_iterator = iter(self)
        else:
            i, current_branch = hint_previous_branch
            current_branch = current_branch.copy()
            if is_box_index:
                box_counter = self.num_boxes_up_to(i)
            current_iterator = self.__iter__(start=i)  # type: ignore
        # traverse tree
        # store/stack how many boxes on this level are left to go up again
        while is_box_index or i < index:
            current_refinement = next(current_iterator)
            if current_refinement == self.d_zeros:
                if is_box_index:
                    box_counter += 1
                    if box_counter > index:
                        break
                current_branch.advance_branch()
            else:
                current_branch.grow_branch(current_refinement)
            i += 1
        return current_branch, current_iterator

    # now a collection of branch-based family-finding functions
    def get_parent(self, child_branch: Branch) -> tuple[int, Iterator]:
        """Find the parent index of a lost child node by advancing and
        growing a new branch until it matches the child's branch sufficiently.

        Args:
            child_branch (Branch): the branch belonging to the child index

        Returns:
            tuple[int, Iterator]: the index of the parent and
                the iterator to one-past-the-parent
        """
        current_iterator = iter(self)
        if len(child_branch) < 2:
            return -1, current_iterator
        # have to start from beginning, note when we go down matching twigs
        current_branch = Branch(self._num_dimensions)
        twig_index = 1
        for i in range(len(self)):
            current_refinement = next(current_iterator)
            if current_refinement == self.d_zeros:
                current_branch.advance_branch()
            else:
                current_branch.grow_branch(current_refinement)
            # if this matches the child at the twig index, count up the twig index
            if (
                twig_index < len(child_branch) - 1
                and current_branch[twig_index] == child_branch[twig_index]
            ):
                twig_index += 1
            if twig_index == len(child_branch) - 1 and len(current_branch) == len(
                child_branch
            ):
                # technically the iterator and branch will be too far advanced now,
                # if we needed to return the branch we could prune the last entry
                break
        return i, current_iterator

    def get_ancestry(self, child_branch: Branch) -> list[int]:
        """like get_parent, but returns a sorted list of all ancestors"""
        if len(child_branch) < 2:
            return list()
        # have to start from beginning, note when we go down matching twigs
        ancestry: set[int] = set()
        current_branch = Branch(self._num_dimensions)
        twig_index = 0
        for i, current_refinement in enumerate(self):
            # if this matches the child at the twig index, count up the twig index
            while (
                twig_index < len(current_branch)
                and current_branch[twig_index] == child_branch[twig_index]
            ):
                twig_index += 1
                ancestry.add(i)
                if twig_index == len(child_branch) - 1:
                    return sorted(ancestry)
            if current_refinement == self.d_zeros:
                current_branch.advance_branch()
            else:
                current_branch.grow_branch(current_refinement)
        raise IndexError("Child branch not found in descriptor")

    def get_oldest_sibling(self, younger_branch: Branch) -> tuple[int, Iterator]:
        parent_index, parent_iterator = self.get_parent(younger_branch)
        return parent_index + 1, parent_iterator

    def get_siblings(
        self, hierarchical_index: int, branch_to_index: Optional[Branch] = None
    ) -> list[int]:
        siblings: set[int] = {hierarchical_index}
        if branch_to_index is None:
            branch, descriptor_iterator = self.get_branch(hierarchical_index, False)
        else:
            branch = branch_to_index.copy()
            descriptor_iterator = self.__iter__(start=hierarchical_index)  # type: ignore
        if len(branch) < 2:
            # we are at the root
            return list(siblings)
        total_num_siblings = 1 << branch[-1].level_increment.count()
        num_older_siblings = total_num_siblings - branch[-1].count_to_go_up
        if num_older_siblings > 0:
            assert not branch_to_index
            running_index, descriptor_iterator = self.get_oldest_sibling(branch)
            siblings.add(running_index)
        else:
            running_index = hierarchical_index

        for _ in range(total_num_siblings - 1):
            current_refinement = next(descriptor_iterator)
            assert current_refinement == self[running_index]
            _, num_patches_skipped = self.skip_to_next_neighbor(
                descriptor_iterator, current_refinement
            )
            running_index += num_patches_skipped
            siblings.add(running_index)

        assert len(siblings) == total_num_siblings
        return sorted(siblings)

    def get_children(
        self, parent_index: int, branch_to_parent: Optional[Branch] = None
    ):
        if self.is_box(parent_index):
            return []
        first_child_index = parent_index + 1
        if branch_to_parent is not None:
            branch_to_first_child = branch_to_parent.copy()
            branch_to_first_child.grow_branch(self[parent_index])
        else:
            branch_to_first_child = None
        return self.get_siblings(first_child_index, branch_to_first_child)

    def get_child_ranges(self, parent_index: int) -> list[tuple[int, int]]:
        """Return (start, end) patch-index ranges for each direct child of parent_index."""
        num_children = get_num_children_from_refinement(self[parent_index])
        if num_children == 0:
            return []
        ranges = []
        current = parent_index + 1
        it = self.__iter__(start=current)  # type: ignore
        for _ in range(num_children):
            child_ref = next(it)
            _, size = self.skip_to_next_neighbor(it, child_ref)
            ranges.append((current, current + size))
            current += size
        return ranges

    def get_level(self, index: int, is_box_index: bool = True) -> npt.NDArray[np.int8]:
        current_branch, _ = self.get_branch(index, is_box_index)
        found_level = get_level_from_branch(current_branch)
        return found_level

    def skip_to_next_neighbor(
        self, descriptor_iterator: Iterator, current_refinement: ba.frozenbitarray
    ) -> tuple[int, int]:
        """Advances the iterator until it points to the end of the current patch and all its children,
        returns the number of boxes or patches it skipped -- including the current one.
        """
        # sweep to the next patch on the same level
        # = count ones and balance them against found boxes
        added_box_index = 0
        added_hierarchical_index = 1
        boxes_to_close = get_num_children_from_refinement(current_refinement)
        if boxes_to_close == 0:
            added_box_index = 1
        else:
            while boxes_to_close > 0:
                next_refinement = next(descriptor_iterator)
                added_hierarchical_index += 1
                boxes_to_close -= 1
                if next_refinement.count() == 0:
                    added_box_index += 1
                else:
                    boxes_to_close += 1 << next_refinement.count()

        return added_box_index, added_hierarchical_index

    def level_iterator(self):
        for current_branch, _ in branch_generator(self):
            yield get_level_from_branch(current_branch)


def branch_generator(descriptor: RefinementDescriptor):
    current_branch = Branch(descriptor.get_num_dimensions())
    i = 0
    for refinement in descriptor:
        yield current_branch, refinement

        i += 1
        if i == len(descriptor):
            return
        if refinement == descriptor.d_zeros:
            current_branch.advance_branch()
        else:
            current_branch.grow_branch(refinement)


def validate_descriptor(descriptor: RefinementDescriptor):
    if len(descriptor._data) % descriptor._num_dimensions != 0:
        raise DyadaInvalidDescriptorError("Uneven number of bits in descriptor")
    try:
        branch, _ = descriptor.get_branch(len(descriptor) - 1, False)
    except IndexError as e:
        raise DyadaInvalidDescriptorError(
            "Descriptor does not form a valid omnitree"
        ) from e
    if not (len(branch) > 0) or any(twig.count_to_go_up != 1 for twig in branch):
        raise DyadaInvalidDescriptorError("Descriptor not fully traversed")

    return True


def int8_ndarray_from_iterable(iterable) -> npt.NDArray[np.int8]:
    return np.fromiter(iterable, dtype=np.int8, count=len(iterable))


def get_level_from_branch(branch: Branch) -> np.ndarray:
    num_dimensions = len(branch[0].level_increment)
    found_level = np.array([0] * num_dimensions, dtype=np.int8)
    for level_count in range(1, len(branch)):
        found_level += int8_ndarray_from_iterable(branch[level_count].level_increment)
    return found_level


def hierarchical_to_box_index_mapping(
    hierarchical_mapping: list[set[int]],
    key_descriptor: RefinementDescriptor,
    value_descriptor: RefinementDescriptor,
) -> list[set[int]]:
    box_mapping = [
        set(
            value_descriptor.to_box_index(new_index)
            for new_index in new_indices
            if value_descriptor.is_box(new_index)
        )
        for old_index, new_indices in enumerate(hierarchical_mapping)
        if key_descriptor.is_box(old_index)
    ]
    return box_mapping


def find_uniqueness_violations(
    descriptor: RefinementDescriptor,
) -> list[set[int]]:
    """
    Find the tuples of indices where the uniqueness condition is violated.
    This is the case when all children of a box have a refinement (==1) where
    the parent box has no refinement (==0).
    """
    violations: list[set[int]] = []
    children_refinement_stack: list[tuple[int, ba.bitarray]] = []
    # iterate the descriptor from the end to the beginning
    for i in range(len(descriptor) - 1, -1, -1):
        if descriptor[i].count() > 0:
            num_children = 2 ** descriptor[i].count()
            # pop the children from the stack
            children = []
            for _ in range(num_children):
                children.append(children_refinement_stack.pop())

            all_children_have_refinement = children[0][1]
            for child in children[1:]:
                all_children_have_refinement &= child[1]
            if (all_children_have_refinement & ~descriptor[i]).count() > 0:
                # uniqueness condition violated, add to violations
                violations.append({i, *[child[0] for child in children]})

        # put on the stack
        children_refinement_stack.append((i, descriptor[i]))
    return violations



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later
from dataclasses import dataclass, field

from dyada.descriptor import RefinementDescriptor, Branch
from dyada.discretization import Discretization
from dyada.linearization import location_code_from_branch
from dyada.markers import (
    MarkersType,
    MarkersMapProxyType,
    get_next_largest_markered_index,
)


def merge_mappings(
    first_mapping: list[set[int]],
    second_mapping: list[set[int]],
) -> list[set[int]]:
    # if either mapping is empty, return the other one
    if not first_mapping:
        return second_mapping
    if not second_mapping:
        return first_mapping
    # merge the mappings
    merged_mapping: list[set[int]] = [set() for _ in range(len(first_mapping))]
    for k, v in enumerate(first_mapping):
        for v_i in v:
            merged_mapping[k] |= second_mapping[v_i]
    return merged_mapping


@dataclass
class _BranchCorrectionState:
    index_mapping: list[set[int]]

    unmodified_branch: Branch
    initial_branch_depth: int

    old_ancestry: list[int]
    old_ancestry_exact: list[bool]

    index_reverse_replace_map: dict[int, dict[int, set[int]]] = field(
        default_factory=dict
    )  # TODO easier type
    leaves_to_forget_except: list[tuple[int, int]] = field(default_factory=list)

    index_to_consider: int = 0


def correct_index_mapping(
    index_mapping: list[set[int]],
    old_discretization: Discretization,
    new_discretization: Discretization,
    markers: MarkersType | MarkersMapProxyType,
) -> None:
    old_descriptor = old_discretization.descriptor
    one_after_last_considered_index = 0

    while one_after_last_considered_index < len(old_descriptor):
        marked_ancestor_index = get_next_largest_markered_index(
            markers, one_after_last_considered_index
        )
        if marked_ancestor_index == -1:
            break

        state = _initialize_branch_state_with_dataclass(
            index_mapping=index_mapping,
            old_descriptor=old_descriptor,
            marked_ancestor_index=marked_ancestor_index,
        )

        one_after_last_considered_index = _walk_branch(
            state=state,
            old_discretization=old_discretization,
            new_discretization=new_discretization,
        )

        _forget_self_remembering_leaves(
            index_mapping=index_mapping,
            marked_ancestor_index=marked_ancestor_index,
            one_after_last_considered_index=one_after_last_considered_index,
            leaves_to_forget_except=state.leaves_to_forget_except,
        )


def _initialize_branch_state_with_dataclass(
    index_mapping: list[set[int]],
    old_descriptor: RefinementDescriptor,
    marked_ancestor_index: int,
) -> _BranchCorrectionState:
    unmodified_branch, _ = old_descriptor.get_branch(
        marked_ancestor_index, is_box_index=False
    )
    initial_branch_depth = len(unmodified_branch)

    old_ancestry = old_descriptor.get_ancestry(unmodified_branch)
    assert len(old_ancestry) == initial_branch_depth - 1

    old_ancestry.append(marked_ancestor_index)
    old_ancestry_exact = [True] * len(old_ancestry)

    return _BranchCorrectionState(
        index_mapping=index_mapping,
        unmodified_branch=unmodified_branch,
        initial_branch_depth=initial_branch_depth,
        old_ancestry=old_ancestry,
        old_ancestry_exact=old_ancestry_exact,
        index_to_consider=marked_ancestor_index,
    )


def _walk_branch(
    state: _BranchCorrectionState,
    old_discretization: Discretization,
    new_discretization: Discretization,
) -> int:
    old_descriptor = old_discretization.descriptor
    while True:
        if old_descriptor.is_box(state.index_to_consider):
            _apply_reverse_replacements(state)

            try:
                _advance_branch_state(state)
            except IndexError:
                return state.index_to_consider + 1
        else:
            state.unmodified_branch.grow_branch(old_descriptor[state.index_to_consider])

        state.index_to_consider += 1

        matching_index, exact_match = _find_matching_new_index(
            state.unmodified_branch,
            old_discretization,
            new_discretization,
        )

        state.old_ancestry.append(state.index_to_consider)
        state.old_ancestry_exact.append(exact_match)

        _handle_index_replacement(
            state=state,
            old_descriptor=old_descriptor,
            new_descriptor=new_discretization.descriptor,
            matching_index=matching_index,
            exact_match=exact_match,
        )


def _handle_index_replacement(
    state: _BranchCorrectionState,
    old_descriptor: RefinementDescriptor,
    new_descriptor: RefinementDescriptor,
    matching_index: int,
    exact_match: bool,
):
    idx = state.index_to_consider

    should_be_replaced = {i for i in state.index_mapping[idx] if i < matching_index}

    if exact_match:
        depth = len(state.unmodified_branch) - 1
        state.index_reverse_replace_map[depth] = {matching_index: should_be_replaced}

        if old_descriptor.is_box(idx) and new_descriptor.is_box(matching_index):
            state.leaves_to_forget_except.append((idx, matching_index))
    else:
        state.index_mapping[idx].add(matching_index)
        state.index_mapping[idx].difference_update(should_be_replaced)


def _apply_reverse_replacements(state: _BranchCorrectionState) -> None:
    for to_modify_depth in range(
        state.initial_branch_depth - 1,
        len(state.unmodified_branch),
    ):
        to_modify_index = state.old_ancestry[to_modify_depth]

        youngest, oldest = _compute_influence_range(
            to_modify_depth,
            state.initial_branch_depth,
            state.old_ancestry_exact,
            len(state.unmodified_branch),
        )

        for depth, replacements in state.index_reverse_replace_map.items():
            if oldest <= depth <= youngest:
                for replace_index, to_replace in replacements.items():
                    if to_replace & state.index_mapping[to_modify_index]:
                        state.index_mapping[to_modify_index].add(replace_index)
                        state.index_mapping[to_modify_index].difference_update(
                            to_replace
                        )


def _advance_branch_state(state: _BranchCorrectionState) -> None:
    state.unmodified_branch.advance_branch(state.initial_branch_depth)

    depths_to_remove = [
        d for d in state.index_reverse_replace_map if d >= len(state.unmodified_branch)
    ]
    for d in depths_to_remove:
        state.index_reverse_replace_map.pop(d)

    del state.old_ancestry[len(state.unmodified_branch) - 1 :]
    del state.old_ancestry_exact[len(state.unmodified_branch) - 1 :]


def _compute_influence_range(
    to_modify_depth: int,
    initial_branch_depth: int,
    old_ancestry_exact: list[bool],
    branch_length: int,
) -> tuple[int, int]:
    if old_ancestry_exact[to_modify_depth]:
        return to_modify_depth, to_modify_depth

    oldest = initial_branch_depth - 1
    for d in range(to_modify_depth, -1, -1):
        if old_ancestry_exact[d]:
            oldest = d
            break

    youngest = to_modify_depth
    for d in range(to_modify_depth + 1, branch_length):
        if old_ancestry_exact[d]:
            youngest = d - 1
            break

    return youngest, oldest


def _forget_self_remembering_leaves(
    index_mapping: list[set[int]],
    marked_ancestor_index: int,
    one_after_last_considered_index: int,
    leaves_to_forget_except: list[tuple[int, int]],
) -> None:
    for index in range(marked_ancestor_index, one_after_last_considered_index):
        for old_leaf, new_leaf in leaves_to_forget_except:
            if index != old_leaf:
                index_mapping[index].discard(new_leaf)


def _find_matching_new_index(
    unmodified_branch: Branch,
    old_discretization: Discretization,
    new_discretization: Discretization,
) -> tuple[int, bool]:
    old_location_code = location_code_from_branch(
        unmodified_branch, old_discretization._linearization
    )
    try:
        return (
            new_discretization.get_index_from_location_code(
                old_location_code, get_box=False  # type: ignore
            ),
            True,
        )
    except Discretization.NoExactMatchError as e:
        return e.closest_match, False
